from typing import Any, Optional

from nicegui import app, ui, elements
import numpy as np
import pyperclip

from .utils import *
from .map_render import SettingsPanel, MapScene
from ..utils import parse_number, parse_range, parse_xy_range, pretty_fraction, pretty_list
from .. import synth_format, movement, pattern_generation

def _icon_scale(val: str|None, icons: dict) -> str:
    if not val:
        return icons.get(0, "close")
    try:
        v = parse_number(val)
    except ValueError:
        return "error"
    if v > 1:
        return icons.get(1, "add")
    if v < 1:
        return icons.get(-1, "remove")
    return icons.get(0, "close")

def _mark_inputs(on: bool, *inp: ui.input) -> None:
    for i in inp:
        if on:
            i.props('bg-color="grey-6"')
        else:
            i.props(remove='bg-color="grey-6"')

def _register_marking(bt: ui.button, *inp: ui.input) -> ui.button:
    bt.on("pointerenter", lambda _: _mark_inputs(True, *inp))
    bt.on("pointerleave", lambda _: _mark_inputs(False, *inp))
    return bt

def _clear_button(*inp: ui.input) -> ui.button:
    def _clear_all() -> None:
        for o in inp:
            o.set_value(getattr(o, "default_value", "0"))
    bt = ui.button("Reset", icon="delete", color="negative", on_click=_clear_all)
    bt.props("outline size=sm align=left").classes("w-full")
    _register_marking(bt, *inp)
    return bt

def _find_first(types: tuple[str, ...] = synth_format.ALL_TYPES) -> Optional[tuple[str, "numpy array (3)"]]:
    with safe_clipboard_data(use_original=False, write=False) as d:
        return d.find_first()


def _find_first_pair(types: tuple[str, ...] = synth_format.ALL_TYPES) -> Optional[tuple[str, "numpy array (3)", "numpy array (3)"]]:
    with safe_clipboard_data(use_original=False, write=False) as d:
        first_t: Optional[str] = None
        first: Optional["numpy array (3+)"] = None
        second: Optional["numpy array (3+)"] = None
        for t in types:
            ty_objs = sorted(d.get_object_dict(t).items())
            if len(ty_objs) < 2:
                continue
            _, ty_first = ty_objs[0]
            if second is None or ty_first[0,2] < second[2]:
                first_t = t
                first = ty_first[0]
                _, ty_second = ty_objs[1]
                second = ty_second[0]
        if first_t is None:
            return None
        return first_t, first, second

def _stack(
    d: synth_format.DataContainer, count: int, pivot: tuple[float, float, float],
    offset: tuple[float, float, float], scale: tuple[float, float, float], rotation: float, wall_rotation: float, outset: float,
    random_ranges_offset: list[tuple[tuple[float, float], tuple[float, float]]]|None, random_step_offset: tuple[float, float]|None,
    random_ranges_angle: list[tuple[float, float]]|None, random_step_angle: float|None,
    random_ranges_scale: list[tuple[float, float]]|None, random_step_scale: float|None,
):
    pivot_np = np.array(pivot)
    stacking = d.filtered()  # deep copy
    rng = np.random.default_rng()
    for _ in range(count):
        if scale != [1,1,1]:
            stacking.apply_for_all(movement.scale, scale_3d=scale, pivot=pivot_np)
        if rotation:
            stacking.apply_for_all(movement.rotate, angle=rotation, pivot=pivot_np)
        if wall_rotation:
            stacking.apply_for_walls(movement.rotate, angle=wall_rotation, relative=True)
        stacking.apply_for_all(movement.offset, offset_3d=offset)
        if outset:
            stacking.apply_for_all(movement.outset, outset_scalar=outset, pivot=pivot_np)
        if random_ranges_offset is not None or random_ranges_angle is not None or random_ranges_scale is not None:
            tmp = stacking.filtered()  # deep copy
            if random_ranges_offset is not None:
                if len(random_ranges_offset) == 1:
                    area = np.array(random_ranges_offset[0])
                else:
                    areas = np.array([
                        max(x_max-x_min, 0.01)*max(y_max-y_min, 0.01)  # area, where 0-width axes are counted as 0.01 for numerical stability
                        for (x_min, y_min), (x_max, y_max) in random_ranges_offset
                    ])
                    area = rng.choice(random_ranges_offset, p=areas/sum(areas))
                if random_step_offset is not None:
                    xy_min, xy_max = area
                    random_offset = [random_step_offset[axis] * rng.integers(round(xy_min[axis]/random_step_offset[axis]), round(xy_max[axis]/random_step_offset[axis]), endpoint=True) for axis in (0,1)]
                else:
                    random_offset = pattern_generation.random_xy(1, area[0], area[1])[0]
                tmp.apply_for_all(movement.offset, offset_3d=[random_offset[0], random_offset[1], 0])
            if random_ranges_angle is not None:
                if len(random_ranges_angle) == 1:
                    ang_area = random_ranges_angle[0]
                else:
                    ang_areas = np.array([
                        max(a_max-a_min, 0.01)  # ang_area, where 0-width axes are counted as 0.01 for numerical stability
                        for a_max, a_min in random_ranges_angle
                    ])
                    ang_area = rng.choice(random_ranges_angle, p=ang_areas/sum(ang_areas))
                if random_step_angle is not None:
                    random_rotation = random_step_angle * rng.integers(round(ang_area[0]/random_step_angle), round(ang_area[1]//random_step_angle), endpoint=True)
                else:
                    random_rotation = rng.uniform(ang_area[0], ang_area[1])
                tmp.apply_for_all(movement.rotate, angle=random_rotation, pivot=pivot_np)
            if random_ranges_scale is not None:
                if len(random_ranges_scale) == 1:
                    scale_area = random_ranges_scale[0]
                else:
                    scale_areas = np.array([
                        max(a_max-a_min, 0.01)  # ang_area, where 0-width axes are counted as 0.01 for numerical stability
                        for a_max, a_min in random_ranges_scale
                    ])
                    scale_area = rng.choice(random_ranges_scale, p=scale_areas/sum(scale_areas))
                if random_step_scale is not None:
                    random_scale = random_step_scale * rng.integers(round(scale_area[0]/random_step_scale), round(scale_area[1]//random_step_scale), endpoint=True)
                else:
                    random_scale = rng.uniform(scale_area[0], scale_area[1])
                tmp.apply_for_all(movement.scale, scale_3d=np.array((random_scale, random_scale, 1)), pivot=pivot_np)
            d.merge(tmp)
        else:
            d.merge(stacking)

def make_input(label: str, value: str|float, storage_id: str, **kwargs) -> SMHInput:
    default_kwargs: dict[str, str|int] = {"tab_id": "stacking", "width": 24}
    return SMHInput(storage_id=storage_id, label=label, default_value=value, **(default_kwargs|kwargs))


def _stacking_tab() -> None:
    preview_scene: MapScene|None = None
    with ui.row():
        with ui.card():
            ui.label("Pivot")
            pivot_x = make_input("X", "0", "pivot_x", suffix="sq", negate_icons={1: "east", -1: "west"})
            pivot_y = make_input("Y", "0", "pivot_y", suffix="sq", negate_icons={1: "north", -1: "south"})

            @handle_errors
            def _pick_pivot():
                result = _find_first(types=synth_format.NOTE_TYPES)
                if result is None:
                    raise PrettyError(msg="No note found!")
                first_t, first = result
                pivot_x.set_value(pretty_fraction(first[0]))
                pivot_y.set_value(pretty_fraction(first[1]))
                info(f"Set pivot to first note ({first_t} hand{'s' if first_t == 'both' else ''})")
            with _register_marking(ui.button("Pivot", icon="colorize", on_click=_pick_pivot).props("outline size=sm align=left").classes("w-full"), pivot_x, pivot_y):
                ui.tooltip("Place pivot at first note")
            _clear_button(pivot_x, pivot_y)
        with ui.card():
            ui.label("Offset")
            offset_x = make_input("X", "0", "offset_x", suffix="sq", negate_icons={1: "east", -1: "west"})
            offset_y = make_input("Y", "0", "offset_y", suffix="sq", negate_icons={1: "north", -1: "south"})
            @handle_errors
            def _pick_offset():
                tfs = _find_first_pair()
                if tfs is None:
                    raise PrettyError(msg="No object pair found!")
                t, first, second = tfs
                delta = second - first
                offset_x.set_value(pretty_fraction(delta[0]))
                offset_y.set_value(pretty_fraction(delta[1]))
                offset_t.set_value(pretty_fraction(delta[2]))
                if len(delta) >= 5:
                    walls_angle.set_value(str(round((delta[4]+180)%360-180, 4)))
                info(f"Set offset from {t} object pair")
            with ui.button("Offset", icon="colorize", on_click=_pick_offset).props("outline size=sm align=left").classes("w-full") as pick_offset:
                ui.tooltip("Calculate offset between first two objects of the same type")
            _clear_button(offset_x, offset_y)
        with ui.card():
            ui.label("Scale")
            scale_x = make_input("X", "100%", "scale_x", tooltip="Can be given as % or ratio")
            with scale_x.add_slot("prepend"):
                ui.icon("close").classes("rotate-90").bind_name_from(scale_x, "value", lambda v: _icon_scale(v, {1: "unfold_more", -1: "unfold_less"}))
            scale_y = make_input("Y", "100%", "scale_y", tooltip="Can be given as % or ratio")
            with scale_y.add_slot("prepend"):
                ui.icon("close").bind_name_from(scale_y, "value", lambda v: _icon_scale(v, {1: "unfold_more", -1: "unfold_less"}))
            @handle_errors
            def _pick_scale():
                tfs = _find_first_pair()
                if tfs is None:
                    raise PrettyError(msg="No object pair found!")
                t, first, second = tfs
                delta = second - first
                try:
                    p = [pivot_x.parsed_value, pivot_y.parsed_value]
                except ParseInputError as pie:
                    error(f"Error parsing value: {pie.input_id}", pie, data=pie.value)
                    return
                if any(np.isclose(first[:2], p)):
                    error(f"{t} object pair too close to pivot", data={"type": t, "first": first.tolist(), "second": second.tolist(), "pivot": p})
                    return
                s_xy = (second[:2] - p) / (first[:2] - p)
                scale_x.set_value(pretty_fraction(s_xy[0]))
                scale_y.set_value(pretty_fraction(s_xy[1]))
                offset_t.set_value(pretty_fraction(delta[2]))
                if len(delta) >= 5:
                    walls_angle.set_value(str(round((delta[4]+180)%360-180, 4)))
                info(f"Set scale from {t} object pair")
            with ui.button("Scale", icon="colorize", on_click=_pick_scale).props("outline size=sm align=left").classes("w-full mt-auto") as pick_scale:
                ui.tooltip("Calculate scale (XY) between first two objects of the same type")
            _clear_button(scale_x, scale_y)
        with ui.card():
            ui.label("Rotation")
            pattern_angle = make_input("Pattern", "0", "pattern_angle", suffix="°", negate_icons={1: "rotate_left", -1: "rotate_right"})
            outset_amount = make_input("Outset", "0", "outset", suffix="sq", negate_icons={1: "open_in_full", -1: "close_fullscreen"})
            @handle_errors
            def _pick_rot():
                tfs = _find_first_pair()
                if tfs is None:
                    raise PrettyError(msg="No object pair found!")
                t, first, second = tfs
                delta = second - first
                try:
                    p = [pivot_x.parsed_value, pivot_y.parsed_value]
                except ParseInputError as pie:
                    error(f"Error parsing value: {pie.input_id}", pie, data=pie.value)
                    return
                if any(np.isclose(first[:2], p)):
                    error(f"{t} object pair too close to pivot", settings={"pivot": p}, data={"type": t, "first": first.tolist(), "second": second.tolist()})
                    return
                first = first[:2] - p
                second = second[:2] - p
                ang = np.degrees(np.arctan2(first[0], first[1])) - np.degrees(np.arctan2(second[0], second[1]))
                pattern_angle.set_value(pretty_fraction(ang))
                outset_amount.set_value(pretty_fraction(np.sqrt(second.dot(second)) - np.sqrt(first.dot(first))))
                offset_t.set_value(pretty_fraction(delta[2]))
                if len(delta) >= 5:
                    walls_angle.set_value(str(round((delta[4]+360-ang+180)%360-180, 4)))
                info(f"Set rotation and outset from {t} object pair")
            with ui.button("Rotate", icon="colorize", on_click=_pick_rot).props("outline size=sm align=left").classes("w-full") as pick_rot:
                ui.tooltip("Calculate Rotation and Outset between first two objects of the same type")
            _clear_button(pattern_angle, outset_amount)
        with ui.card():
            ui.label("All")
            walls_angle = make_input("Wall rotation", "0", "walls_angle", suffix="°", negate_icons={1: "rotate_left", -1: "rotate_right"})
            offset_t = make_input("Time", "1/16", "offset_t", suffix="b", negate_icons={1: "fast_forward", -1: "fast_rewind"})
            @handle_errors
            def _pick_spiral():
                tfs = _find_first_pair(synth_format.WALL_TYPES)
                if tfs is None:
                    raise PrettyError(msg="No object pair found!")
                t, first, second = tfs
                ang = (second[4] - first[4])
                divisor = np.sin(np.radians(ang/2))
                if divisor == 0:  # avoid division by 0 when angles match
                    error(f"Wall pair ({t}) have matching angle, cannot determine spiral", data={"first": first.tolist(), "second": second.tolist()})
                    return
                # calculate pivot naively
                p = first[:3] + movement.rotate((second - first)[:3]/2, 90 - ang/2) / divisor
                pivot_x.set_value(pretty_fraction(p[0]))
                pivot_y.set_value(pretty_fraction(p[1]))

                pattern_angle.set_value(str(round((ang+180)%360-180, 4)))
                offset_t.set_value(pretty_fraction(second[2] - first[2]))
            
                # reset the rest
                offset_x.set_value(offset_x.default_value)
                offset_y.set_value(offset_y.default_value)
                scale_x.set_value(scale_x.default_value)
                scale_y.set_value(scale_y.default_value)
                outset_amount.set_value(outset_amount.default_value)
                walls_angle.set_value(walls_angle.default_value)

                info(f"Calculated spiral from {t} wall pair")
            all_values = (
                pivot_x, pivot_y,
                offset_x, offset_y,
                scale_x, scale_y,
                pattern_angle, walls_angle,
                offset_t, outset_amount,
            )
            with _register_marking(ui.button("Spiral", icon="colorize", on_click=_pick_spiral).props("outline size=sm align=left").classes("w-full"), *all_values) as pick_spiral:
                ui.tooltip('Calculates "intuitive" spiral using two walls of the same type')
            _clear_button(*all_values)
        _register_marking(pick_offset, offset_x, offset_y, offset_t, walls_angle)
        _register_marking(pick_scale, scale_x, scale_y, offset_t, walls_angle)
        _register_marking(pick_rot, pattern_angle, outset_amount, offset_t, walls_angle)

        @handle_errors
        def _do_stack(count_mode: str):
            o_t = offset_t.parsed_value
            if not o_t:
                raise PrettyError("Time offset must not be 0", data=offset_t.value)
            p = (pivot_x.parsed_value, pivot_y.parsed_value, 0)
            o = (offset_x.parsed_value, offset_y.parsed_value, o_t)
            s = (scale_x.parsed_value, scale_y.parsed_value, 1)
            r = pattern_angle.parsed_value
            wr = walls_angle.parsed_value
            outset = outset_amount.parsed_value
            # random offset
            try:
                if not random_offset.value:  # empty string or None
                    random_ranges_offset = random_step_offset = None
                elif "@" in random_offset.value:
                    ranges, step = random_offset.value.rsplit("@", 1)
                    random_ranges_offset = [parse_xy_range(r) for r in ranges.split(";")]
                    if "," in step:
                        x, y = step.split(",",1)
                        random_step_offset = (parse_number(x), parse_number(y))
                    else:
                        xy = parse_number(step)
                        random_step_offset = (xy, xy)
                else:
                    random_ranges_offset = [parse_xy_range(r) for r in random_offset.value.split(";")]
                    random_step_offset = None
            except ValueError as ve:
                raise ParseInputError(input_id="stacking_random_offset", value=random_offset.value, exc=ve) from ve
            # random angle
            try:
                if not random_angle.value:  # empty string or None
                    random_ranges_angle = random_step_angle = None
                elif "@" in random_angle.value:
                    ranges, step = random_angle.value.rsplit("@", 1)
                    random_ranges_angle = [parse_range(r) for r in ranges.split(";")]
                    random_step_angle = parse_number(step)
                else:
                    random_ranges_angle = [parse_range(r) for r in random_angle.value.split(";")]
                    random_step_angle = None
            except ValueError as ve:
                raise ParseInputError(input_id="stacking_random_angle", value=random_angle.value, exc=ve) from ve
            # random scale
            try:
                if not random_scale.value:  # empty string or None
                    random_ranges_scale = random_step_scale = None
                elif "@" in random_scale.value:
                    ranges, step = random_scale.value.rsplit("@", 1)
                    random_ranges_scale = [parse_range(r) for r in ranges.split(";")]
                    random_step_scale = parse_number(step)
                else:
                    random_ranges_scale = [parse_range(r) for r in random_scale.value.split(";")]
                    random_step_scale = None
            except ValueError as ve:
                raise ParseInputError(input_id="stacking_random_scale", value=random_scale.value, exc=ve) from ve
            try:
                with safe_clipboard_data(use_original=True, realign_start=False) as d:  # type: synth_format.ClipboardDataContainer
                    if count_mode == "count":
                        c = int(count.parsed_value)
                    elif count_mode == "duration":
                        c = int(duration.parsed_value / o_t)
                    elif count_mode == "fill":
                        c = int(d.selection_length / o_t)
                    _stack(
                        d=d, count=c, pivot=p,
                        offset=o, scale=s, rotation=r, wall_rotation=wr, outset=outset,
                        random_ranges_offset=random_ranges_offset, random_step_offset=random_step_offset,
                        random_ranges_angle=random_ranges_angle, random_step_angle=random_step_angle,
                        random_ranges_scale=random_ranges_scale, random_step_scale=random_step_scale,
                    )
            except PrettyError:
                raise
            except Exception as exc:
                raise PrettyError(
                    msg=f"Error executing stack",
                    exc=exc,
                    context={
                        "count": c, "pivot": p,
                        "offset": o, "scale": s, "rotation": r, "wall_rotation": wr, "outset": outset,
                        "random_ranges_offset": random_ranges_offset, "random_step_offset": random_step_offset,
                        "random_ranges_angle": random_ranges_angle, "random_step_angle": random_step_angle,
                    },
                ) from exc
            counts = d.get_counts()
            info(
                f"Completed stack",
                caption=pretty_list([f"{counts[t]['total']} {t if counts[t]['total'] != 1 else t.rstrip('s')}" for t in ("notes", "rails", "rail_nodes", "walls")]),
            )
            if preview_scene is not None:
                preview_settings = sp.parse_settings()
                preview_scene.render(d, preview_settings)

        with ui.card():
            ui.label("Execute stack")
            with ui.button("Fill", icon="format_color_fill", on_click=lambda _: _do_stack("fill")).props("rounded").classes("w-full"):
                ui.tooltip("Stack until end of selection")
            ui.separator()
            with ui.row():
                count = make_input("Count", 15, "count", suffix="x", width=12)
                ui.button(icon="play_arrow", on_click=lambda _: _do_stack("count")).props("rounded").classes("w-10 mt-auto")
            ui.separator()
            with ui.row():
                duration = make_input("Duration", 1, "duration", suffix="b", width=12)
                ui.button(icon="play_arrow", on_click=lambda _: _do_stack("duration")).props("rounded").classes("w-10 mt-auto")
        with ui.card():
            ui.label("Substeps")
            subdiv_values = (
                offset_x, offset_y,
                scale_x, scale_y,
                pattern_angle, walls_angle,
                offset_t, outset_amount,
            )
            @handle_errors
            def _subdivide(substeps: float) -> None:
                for v in (offset_x, offset_y, pattern_angle, walls_angle,offset_t, outset_amount):
                    v.set_value(pretty_fraction(v.parsed_value/substeps))
                for v in (scale_x, scale_y):
                    v.set_value(f"{v.parsed_value**(1/substeps):.1%}")
            subdiv = make_input("Substeps", "2", "subdiv", tooltip="Number of substeps (should be >1)", suffix="x")
            with subdiv.add_slot("prepend"):
                ui.icon("density_medium")
            _register_marking(ui.button("Subdivide", on_click=lambda _: _subdivide(subdiv.parsed_value if subdiv.parsed_value else 1)).props("outline size=sm").classes("w-full"), *subdiv_values).tooltip("Divide current step into substeps")
            _register_marking(ui.button("Combine", on_click=lambda _: _subdivide((1/subdiv.parsed_value) if subdiv.parsed_value else 1)).props("outline size=sm").classes("w-full"), *subdiv_values).tooltip("Combine multiple steps into one")

        with ui.card():
            with ui.dialog() as random_dialog, ui.card():
                ui.markdown("""
                    This is applied on each copy individually, so the random range will stay centered on the unrandomized stack and not "drift".

                    Ranges are always uniformly distributed. To get non-uniform distributions, you can overlap ranges.

                    |Input|Example|Description|
                    |-|-|-|
                    |`<x>,<y>` |`5,3` |X is chosen randomly between -5 to +5 and Y between -3 and +3, equivalent to "-5:5,-3:3"|
                    |`<min_x>:<max_x>, <min_y>:<max_y>` |`-3:5, -1:1` |X is chosen randomly between -3 to +5 and Y between -1 and +1|
                    |`<range A>; <range B>` |`-10:-5, 1;5:10,1` |Multiple ranges (weighted based on size): Any offset in the two rectangles from -10,-1 to -5,+1 and +5,-1 to +10,+1|
                    |`... @<step_x>,<step_y>` |`-5,0:-3,0; 1,0:2,0 @1,1` |Must be at the end: Random values are always a multiple of the step (1sq here), so ie X will be chosen randomly from -5,-4,-3,+1,+2, while Y is always 0. When just step_x is given, that step is also used for Y|

                    Rotation and scaling works the same, except there is only a single value, not X and Y (ie `5;-2:1@0.5` or `500%;-200%:100%@25%`)
                """)
            with ui.label("Random"):
                with ui.button(icon="help", on_click=random_dialog.open).props('flat text-color=info').classes("w-4 h-4 text-xs cursor-help"):
                    ui.tooltip("Show range input format help")
            # custom input format, don't use SMHInput
            random_offset = ui.input("XY Offset Range", value="").props('dense suffix="sq"').classes("w-24").bind_value(app.storage.user, "stacking_random_offset")
            random_offset.default_value = ""
            random_angle = ui.input("Rotation Range", value="").props('dense suffix="°"').classes("w-24").bind_value(app.storage.user, "stacking_random_angle")
            random_angle.default_value = ""
            random_scale = ui.input("Scale Range", value="").props('dense suffix="°"').classes("w-24").bind_value(app.storage.user, "stacking_random_scale")
            random_scale.default_value = ""
            _clear_button(random_offset, random_angle, random_scale)

    with ui.card():
        @handle_errors
        def _soft_refresh():
            try:
                data = synth_format.ClipboardDataContainer.from_json(read_clipboard())
            except:
                # fall back to empty data on error
                data = synth_format.DataContainer()
            preview_settings = sp.parse_settings()
            if preview_scene is None:
                draw_preview_scene.refresh()
            if preview_scene is not None:
                preview_scene.render(data, preview_settings)
        with ui.row():
            with ui.row():
                ui.label("Clipboard Preview").classes("my-auto")
                with ui.button(icon="sync", on_click=_soft_refresh, color="positive").props("outline"):
                    ui.tooltip("Preview current clipboard")
            with ui.expansion("Preview Settings", icon="palette").props("dense"):
                sp = SettingsPanel()
                ui.separator()
                with ui.row():
                    ui.icon("preview", size="3em").tooltip("Change size and scaling of preview")
                    scene_width = make_input("Width", "800", "width", tab_id="preview", suffix="px", tooltip="Width of the preview in px", width=20)
                    scene_height = make_input("Height", "600", "height", tab_id="preview", suffix="px", tooltip="Height of the preview in px", width=20)
                    time_scale = make_input("Time Scale", "64", "time_scale", tab_id="preview", tooltip="Ratio between XY and time", width=20)
                    frame_length = make_input("Frame Length", "16", "frame_length", tab_id="preview", suffix="b", tooltip="Number of beats to draw frames for", width=20)
                apply_button = ui.button("Apply").props("outline")

        @ui.refreshable
        @handle_errors
        def draw_preview_scene():
            nonlocal preview_scene
            w = int(scene_width.parsed_value)
            h = int(scene_height.parsed_value)
            l = int(frame_length.parsed_value)
            t = time_scale.parsed_value
            preview_scene = MapScene(width=w, height=h, frame_length=l, time_scale=t)
            _soft_refresh()
        draw_preview_scene()
        apply_button.on("click", draw_preview_scene.refresh)

stacking_tab = GUITab(
    name="stacking",
    label="Stacking",
    icon="layers",
    content_func=_stacking_tab,
)



from contextlib import contextmanager
from functools import partial
from typing import Any, Callable, Optional, List, Literal

import numpy as np
from nicegui import app, events, ui
import pyperclip

from .utils import *
from .. import movement, pattern_generation, rails, synth_format
from ..utils import parse_number, pretty_list, pretty_fraction

def _safe_inverse(v: float) -> float:
    return 0.0 if v == 0 else 1/v

def _safe_parse_number(v: Any) -> float:
    try:
        return parse_number(v)
    except ValueError:
        return 0.0

def _swap_hands(data: synth_format.DataContainer, **kwargs) -> None:
    data.left, data.right = data.right, data.left

def _change_color(data: synth_format.DataContainer, types: List[str], rail_filter: synth_format.RailFilter|None, new_type: str, **kwargs) -> None:
    # to single type: just merge all dicts
    changed: synth_format.SINGLE_COLOR_NOTES = {}
    for t in types:
        if t in synth_format.NOTE_TYPES and t != new_type:
            if not rail_filter:
                changed |= getattr(data, t)
                setattr(data, t, {})
            else:
                unchanged = {}
                for ti, nodes in getattr(data, t).items():
                    if rail_filter.matches(nodes):
                        changed[ti] = nodes
                    else:
                        unchanged[ti] = nodes
                setattr(data, t, unchanged)
    # existing notes always have priority
    setattr(data, new_type, changed | getattr(data, new_type))

def _space_walls(data: synth_format.DataContainer, interval: float) -> None:
    out: synth_format.WALLS = {}
    for i, (_, w) in enumerate(sorted(data.walls.items())):
        out[i*interval] = (w*[1,1,0,1,1]) + [0,0,i*interval,0,0]
    data.walls = out

def make_input(label: str, value: str|float, storage_id: str, **kwargs) -> SMHInput:
    default_kwargs = {"tab_id": "dashboard"}
    return SMHInput(storage_id=storage_id, label=label, default_value=value, **(default_kwargs|kwargs))

def offset_card(action_btn_cls: Any) -> None:
    with ui.card():
        with ui.label("Offset"):
            wiki_reference("Movement-Options#offset")
        with ui.grid(columns=3):
            for y in (1, 0, -1):
                for x in (-1, 0, 1):
                    if x or y:
                        action_btn_cls(
                            tooltip=f'Offset {("down", "", "up")[y+1]}{" and " if x and y else ""}{("left", "", "right")[x+1]}',
                            icon=f'{("south", "", "north")[y+1]}{"_" if x and y else ""}{("west", "", "east")[x+1]}',
                            apply_func=movement.offset,
                            apply_args=lambda x=x, y=y: dict(offset_3d=np.array([x,y,0])*offset_xy.parsed_value),
                        )
                    else:
                        offset_xy = make_input("X/Y", 1, "offset_xy", suffix="sq")
        ui.separator()
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Offset earlier in time",
                icon="remove",
                apply_func=movement.offset,
                apply_args=lambda: dict(offset_3d=np.array([0,0,-offset_t.parsed_value])),
                color="secondary",
            )
            offset_t = make_input("Time", 1, "dashboard_offset_t", suffix="b")
            action_btn_cls(
                tooltip="Offset later in time",
                icon="add",
                apply_func=movement.offset,
                apply_args=lambda: dict(offset_3d=np.array([0,0,offset_t.parsed_value])),
            )

def scaling_card(action_btn_cls: Any) -> None:
    with ui.card():
        with ui.label("Scaling"):
            wiki_reference("Movement-Options#scaling")
        with ui.grid(columns=3):
            scaleup_label = ui.label("110%").classes("ml-auto h-min bg-primary")
            action_btn_cls(
                tooltip="Scale Y up (taller)",
                icon="unfold_more",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([1,scale_xy.parsed_value,1])),
            )
            action_btn_cls(
                tooltip="Scale XY up (larger)",
                icon="zoom_out_map",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([scale_xy.parsed_value,scale_xy.parsed_value,1])),
            )
            action_btn_cls(
                tooltip="Scale X down (less wide)",
                icon="unfold_less", icon_angle=90,
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([_safe_inverse(scale_xy.parsed_value),1,1])),
                color="secondary",
            )
            scale_xy = make_input("X/Y", "110%", "scale_xy", tooltip="Can be given as % or ratio. If less than 1 (100%), scale up/down are inverted")
            action_btn_cls(
                tooltip="Scale X up (wider)",
                icon="unfold_more", icon_angle=90,
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([scale_xy.parsed_value,1,1])),
            )
            action_btn_cls(
                tooltip="Scale XY down (smaller)",
                icon="zoom_in_map",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([_safe_inverse(scale_xy.parsed_value),_safe_inverse(scale_xy.parsed_value),1])),
                color="secondary",
            )
            action_btn_cls(
                tooltip="Scale Y down (less tall)",
                icon="unfold_less",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([1,_safe_inverse(scale_xy.parsed_value),1])),
                color="secondary",
            )
            with ui.label().classes("mt-auto w-min bg-secondary").bind_text_from(scale_xy, "value", backward=lambda v: f"{_safe_inverse(_safe_parse_number(v)):.1%}"):
                ui.tooltip("This is the exact inverse of the scale up. Percent calculations are weird like that.")
            scaleup_label.bind_text_from(scale_xy, "value", backward=lambda v: f"{_safe_parse_number(v):.1%}")
        ui.separator()
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Scale time down (shorter)",
                icon="compress",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([1,1,_safe_inverse(scale_t.parsed_value)])),
                color="secondary",
            )
            scale_t = make_input("Time", 2, "scale_t", tooltip="Can be given as % or ratio. If less than 1 (100%), scale up/down are inverted")
            action_btn_cls(
                tooltip="Scale time up (longer)",
                icon="expand",
                apply_func=movement.scale,
                apply_args=lambda: dict(scale_3d=np.array([1,1,scale_t.parsed_value])),
            )
            
            action_btn_cls(
                tooltip="Read BPM from clipboard",
                icon="colorize",
                func=lambda data, **kwargs: setattr(scale_bpm, "value", str(data.bpm)),
            ).props("outline")
            scale_bpm = make_input("New BPM", 120, "scale_bpm")
            action_btn_cls(
                tooltip="Change BPM of clipboard (keeps timing)",
                icon="straighten",
                func=lambda data, **kwargs: data.change_bpm(scale_bpm.parsed_value),
                wiki_ref="Pre--and-Post-Processing-Options#change-bpm",
            )

def flatten_mirror_card(action_btn_cls: Any) -> None:
    with ui.card():
        ui.label("Flatten").tooltip("Just scaling, but with 0")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Flatten to Y axis (X=0)",
                icon="vertical_align_center", icon_angle=90,
                apply_func=partial(movement.scale, scale_3d=np.array([0,1,1])),
            )
            action_btn_cls(
                tooltip="Flatten to X axis (Y=0)",
                icon="vertical_align_center",
                apply_func=partial(movement.scale, scale_3d=np.array([1,0,1])),
            )
            action_btn_cls(
                tooltip="Move to pivot (X=Y=0)",
                icon="adjust",
                apply_func=partial(movement.scale, scale_3d=np.array([0,0,1])),
            )
        ui.separator()
        ui.label("Mirror").tooltip("Just scaling, but with -1")
        with ui.row():
            def _do_mirror_axis(data: synth_format.DataContainer, axis: int, **kwargs):
                # work on copy when stacking, else directly on data
                tmp = data.filtered() if mirror_do_stack.value else data
                scale_3d = np.ones((3,))
                scale_3d[axis] = -1
                tmp.apply_for_all(movement.scale, scale_3d=scale_3d, **kwargs)
                if mirror_do_stack.value:
                    interval = mirror_stack_interval.parsed_value
                    if axis != 2 and not interval:
                        raise PrettyError("Stacked copy with interval 0 would just override input.")
                    tmp.apply_for_all(movement.offset, [0,0,interval])
                    data.merge(tmp)
            action_btn_cls(
                tooltip="Mirror X (left<->right)",
                icon="align_horizontal_center",
                func=partial(_do_mirror_axis, axis=0),
            )
            action_btn_cls(
                tooltip="Mirror Y (up<->down)",
                icon="align_vertical_center",
                func=partial(_do_mirror_axis, axis=1),
            )
            action_btn_cls(
                tooltip="Mirror time (reverse). Enabling realigning (orange slider in the top left) is recommended.",
                icon="fast_rewind",
                func=partial(_do_mirror_axis, axis=2),
            )
        with ui.row():
            mirror_angle = make_input("Custom Mirror Angle", 45, "mirror_angle", suffix="°", tooltip="Angle of the mirror line. 0: --, ±90: |, +45: /, -45: \\", width=28)
            def _do_mirror_custom(data: synth_format.DataContainer, **kwargs):
                    # work on copy when stacking, else directly on data
                    tmp = data.filtered() if mirror_do_stack.value else data
                    # subtract rotation, mirror, add back rotation
                    tmp.apply_for_all(movement.rotate, angle=-mirror_angle.parsed_value, **kwargs)
                    tmp.apply_for_all(movement.scale, scale_3d=np.array([1,-1,1]), **kwargs)
                    tmp.apply_for_all(movement.rotate, angle=mirror_angle.parsed_value, **kwargs)
                    if mirror_do_stack.value:
                        tmp.apply_for_all(movement.offset, [0,0,mirror_stack_interval.parsed_value])
                        data.merge(tmp)
            custom_mirror_btn = action_btn_cls(
                tooltip="",
                icon=None,
                func=_do_mirror_custom,
            )
            @handle_errors
            def _rotate_mirror_icon() -> None:
                try:
                    ang = mirror_angle.parsed_value
                except ParseInputError:
                    ang = 0
                mirror_icon.style(f"rotate: {90-ang}deg")
            with custom_mirror_btn.add_slot("default"):
                ui.tooltip("Mirror with custom angle. Depending on coordinate mode, the mirror line passes through grid center, object center or pivot")
                mirror_icon = ui.icon("flip")
                _rotate_mirror_icon()
            mirror_angle.on_value_change(_rotate_mirror_icon)
        with ui.row():
            with ui.switch("Stacking", value=False).props('size="xs"').classes("w-28 my-auto").bind_value(app.storage.user, "dashboard_mirror_do_stack") as mirror_do_stack:
                ui.tooltip("Instead of simply mirroring the input, stack the mirrored copy behind it.")
            mirror_stack_interval = make_input("Interval", 0, "mirror_stack_interval", suffix="b", tooltip="Interval for stacked copy. When reversing time, this should either be 0 (to mirror across cursor) or double the pattern length (to mirror from back).").bind_enabled_from(mirror_do_stack, "value")

def rotate_outset_card(action_btn_cls: Any) -> None:
    with ui.card():
        with ui.label("Rotation"):
            wiki_reference("Movement-Options#rotate")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Rotate counterclockwise",
                icon="rotate_left",
                apply_func=movement.rotate,
                apply_args=lambda: dict(angle=rotate_angle.parsed_value),
            )
            rotate_angle = make_input("Angle", 45, "angle", suffix="°")
            action_btn_cls(
                tooltip="Rotate clockwise",
                icon="rotate_right",
                apply_func=movement.rotate,
                apply_args=lambda: dict(angle=-rotate_angle.parsed_value),
            )
        ui.separator()

        with ui.label("Outset"):
            wiki_reference("Movement-Options#outset")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Inset (towards center/pivot)",
                icon="close_fullscreen",
                apply_func=movement.outset,
                apply_args=lambda: dict(outset_scalar=-outset_amount.parsed_value),
                color="secondary",
            )
            outset_amount = make_input("Amount", 1, "outset", suffix="sq")
            action_btn_cls(
                tooltip="Outset (away from center/pivot)",
                icon="open_in_full",
                apply_func=movement.outset,
                apply_args=lambda: dict(outset_scalar=outset_amount.parsed_value),
            )

        ui.separator()
        with ui.label("Create parallel"):
            wiki_reference("Pre--and-Post-Processing-Options#create-parallel-patterns")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Create parallel crossovers",
                icon="shuffle",
                func=lambda data, types, rail_filter, **kwargs: pattern_generation.create_parallel(data, distance=-parallel_distance.parsed_value, types=types, rail_filter=rail_filter),
                color="secondary",
            )
            parallel_distance = make_input("Spacing", 2, "parallel", suffix="sq")
            action_btn_cls(
                tooltip="Create parallel pattern",
                icon="stacked_line_chart",
                func=lambda data, types, rail_filter, **kwargs: pattern_generation.create_parallel(data, distance=parallel_distance.parsed_value, types=types, rail_filter=rail_filter),
            )

def rails_card(action_btn_cls: Any) -> None:
    with ui.card():
        ui.label("Rails and Notes")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Merge sequential rails",
                icon="link",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.merge_sequential_rails, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#merge-rails",
            )
            action_btn_cls(
                tooltip="Split rails at single notes",
                icon="link_off",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.split_rails, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#split-rails",
            )
            action_btn_cls(
                tooltip="Snap notes to rail",
                icon="insights",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.snap_singles_to_rail, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#snap-single-notes-to-rails",
            )

            action_btn_cls(
                tooltip="Rail nodes to notes (delete rail)",
                icon="more_vert",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.rails_to_singles, keep_rail=False, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#convert-rails-into-single-notes",
            )
            action_btn_cls(
                tooltip="Rail nodes to notes (keep rail)",
                icon="more_vert"+"show_chart",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.rails_to_singles, keep_rail=True, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#convert-rails-into-single-notes",
            )
        ui.separator()
        with ui.grid(columns=3):
            rail_distance = make_input("Max-Dist", 1, "rail_distance", suffix="b", tooltip="Maximum distance")
            action_btn_cls(
                tooltip="Connect notes",
                icon="linear_scale", icon_angle=90,
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.connect_singles, max_interval=rail_distance.parsed_value, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#connect-single-notes-into-rails",
            )
            action_btn_cls(
                tooltip="Connect rails",
                icon="add_link",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.merge_rails, max_interval=rail_distance.parsed_value, types=types, rail_filter=rail_filter),
                wiki_ref="Pre--and-Post-Processing-Options#merge-rails",
            )
        ui.separator()
        with ui.row():
            rail_interval = make_input("Interval", "1/16", "rail_interval", suffix="b")
            with ui.switch("From end", value=False).props('dense size="xs"').classes("my-auto").bind_value(app.storage.user, "dashboard_rail_from_back") as rail_from_back:
                ui.tooltip("Start operation from the end instead of the start, so e.g. shorten would cut from start instead of end and extending looks at previous instead of next.")
        with ui.grid(columns=3):
            action_btn_cls(
                tooltip="Split rail at time intervals",
                icon="format_line_spacing"+"link_off",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.segment_rail, max_length=rail_interval.parsed_value*(1-rail_from_back.value*2), types=types, rail_filter=rail_filter),
            )
            action_btn_cls(
                tooltip="Interpolate rail nodes",
                icon="format_line_spacing"+"commit",
                apply_func=partial(rails.interpolate_nodes, mode="spline"),
                apply_for="notes",
                apply_args=lambda: dict(interval=rail_interval.parsed_value*(1-rail_from_back.value*2)),
                wiki_ref="Rail-Options#interpolate",
            )
            action_btn_cls(
                tooltip="Extend level",
                icon="swipe_right_alt" + "horizontal_rule",
                apply_func=rails.extend_level,
                apply_for="notes",
                apply_args=lambda: dict(distance=rail_interval.parsed_value*(1-rail_from_back.value*2)),
            )

            action_btn_cls(
                tooltip="Rail to notestack (delete rail)",
                icon="animation",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.rails_to_notestacks, interval=rail_interval.parsed_value*(1-rail_from_back.value*2), keep_rail=False, types=types, rail_filter=rail_filter),
            )
            action_btn_cls(
                tooltip="Rail to notestack (keep rail)",
                icon="animation"+"show_chart",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.rails_to_notestacks, interval=rail_interval.parsed_value*(1-rail_from_back.value*2), keep_rail=True, types=types, rail_filter=rail_filter),
            )
            action_btn_cls(
                tooltip="Extend directional / straight",
                icon="swipe_right_alt" + "double_arrow",
                apply_func=rails.extend_straight,
                apply_for="notes",
                apply_args=lambda: dict(distance=rail_interval.parsed_value*(1-rail_from_back.value*2)),
            )

            action_btn_cls(
                tooltip="Shorten rail by amount from the end",
                icon="content_cut"+"straighten",
                apply_func=rails.shorten_rail_by,
                apply_for="notes",
                apply_args=lambda: dict(distance=rail_interval.parsed_value*(1-rail_from_back.value*2)),
                wiki_ref="Rail-Options#shorten-rails",
            )
            action_btn_cls(
                tooltip="Shorten rail to given length",
                icon="straighten"+"content_cut",
                apply_func=rails.shorten_rail_to,
                apply_for="notes",
                apply_args=lambda: dict(distance=rail_interval.parsed_value*(1-rail_from_back.value*2)),
            )
            action_btn_cls(
                tooltip="Extend pointing to next",
                icon="swipe_right_alt" + "swipe_right_alt",
                apply_for="notes",
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.extend_to_next, distance=rail_interval.parsed_value*(1-rail_from_back.value*2), types=types, rail_filter=rail_filter),
            )

            action_btn_cls(
                tooltip="Create a both-hands lightning rail that zig-zags between the first rail of one hand and rails of the other hand",
                icon="graphic_eq",
                icon_angle=-45,
                func=lambda data, types, rail_filter, **kwargs: pattern_generation.create_lightning(data=data, interval=rail_interval.parsed_value*(1-rail_from_back.value*2), rail_filter=rail_filter),
            ).classes("text-yellow")

def smoothing_card(action_btn_cls: Any) -> None:
    with ui.card():
        ui.label("Smoothing")
        with ui.row():
            smooth_iterations = make_input("Iterations", "3", "smooth_interations", tooltip="Number of smoothing iterations. Higher number leads to more smoothing.")
            smooth_resistance = make_input("Resist", "1", "smooth_resistance", tooltip="Resistance of original rail against smoothing. Higher values results in smaller changes.")
            action_btn_cls(
                tooltip="Smooth rail nodes",
                icon="switch_access_shortcut",
                icon_angle=90,
                func=lambda data, types, rail_filter, **kwargs: data.apply_for_note_types(rails.reinterpolation_smoothing, 
                    iterations=int(smooth_iterations.parsed_value),
                    resistance=smooth_resistance.parsed_value,
                    singles_mode=smooth_singlesmode.value,
                    types=types,
                    rail_filter=rail_filter,
                ),
            )
        with ui.row().classes("w-full"):
            smooth_singlesmode = ui.radio({"snap":"", "anchor":"", "temp_anchor":""}, value="anchor").props("dense inline").bind_value(app.storage.user, "dashboard_smooth_singlesmode")
            with ui.teleport(f"#c{smooth_singlesmode.id} > div:nth-child(1)"):
                ui.tooltip("Snap single notes to smoothed rail")
                ui.icon("moving", size="sm")
            with ui.teleport(f"#c{smooth_singlesmode.id} > div:nth-child(2)"):
                ui.tooltip("Single notes anchor rail for smoothing")
                ui.icon("anchor", size="sm")
            with ui.teleport(f"#c{smooth_singlesmode.id} > div:nth-child(3)"):
                ui.tooltip("Remove anchors notes after smoothing (if you placed them as tempoary anchors)")
                ui.icon("delete", size="sm")


def color_card(action_btn_cls: Any) -> None:
    with ui.card():
        with ui.label("Color"):
            wiki_reference("Pre--and-Post-Processing-Options#change-note-type")

        action_btn_cls(
            tooltip="Swap Hands (ignores filters)",
            icon="swap_horizontal_circle",
            func=_swap_hands,
            wiki_ref="Pre--and-Post-Processing-Options#swap-hands",
        ).props("outline")
        action_btn_cls(
            tooltip="Change to left hand",
            icon="change_circle",
            func=partial(_change_color, new_type="left"),
            color="cyan",
        )
        action_btn_cls(
            tooltip="Change to right hand",
            icon="change_circle",
            func=partial(_change_color, new_type="right"),
            color="pink",
        )
        action_btn_cls(
            tooltip="Change to single hand",
            icon="change_circle",
            func=partial(_change_color, new_type="single"),
            color="green",
        )
        action_btn_cls(
            tooltip="Change to both hands",
            icon="change_circle",
            func=partial(_change_color, new_type="both"),
            color="amber",
        )

def spiral_spike_card(action_btn_cls: Any) -> None:
    with ui.card():
        ui.label("Spiralize and Spikify")
        with ui.row():
            spiral_angle = make_input("Angle", 45, "spiral_angle", suffix="°", tooltip="Angle between nodes. Choose 180 for zigzag.")
            spiral_start = make_input("Start", 0, "spiral_start", suffix="°", tooltip="Angle of first node: 0=right, 90=up, 180=left, 270/-90=down")
            spiral_radius = make_input("Radius", 1, "spiral_radius", suffix="sq", tooltip="Radius of spiral / Length of spikes")
        with ui.row():
            with ui.switch("Interpolate", value=True).props('size="xs"').classes("my-auto").bind_value(app.storage.user, "dashboard_spiral_do_interpolate") as spiral_do_interpolate:
                ui.tooltip("Interpolate rail before adding spiral/spikes. Enable this for consistent spacing.")
            spiral_interpolation = make_input("Interval", "1/16", "spiral_interpolation", suffix="b", tooltip="Time between spiral nodes/spikes").bind_enabled(spiral_do_interpolate, "value")
        with ui.label("Spiral"):
            wiki_reference("Rail-Options#spiral")
        with ui.row():
            def _add_spiral(nodes: "numpy array (n, 3)", fid_dir: int, direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,) -> "numpy array (n, 3)":
                if spiral_do_interpolate.value:
                    nodes = rails.interpolate_nodes(nodes, mode="spline", interval=spiral_interpolation.parsed_value)
                return pattern_generation.add_spiral(
                    nodes,
                    fidelity=fid_dir*(360*_safe_inverse(spiral_angle.parsed_value) or 1),  # if angle=0, default to 1
                    radius=spiral_radius.parsed_value,
                    start_angle=spiral_start.parsed_value,
                    direction=direction,
                )
            action_btn_cls(
                tooltip="Spiral (counter-clockwise)",
                icon="rotate_left",
                apply_func=partial(_add_spiral, fid_dir=1),
                apply_for="notes",
            ).props("rounded")
            action_btn_cls(
                tooltip="Spiral (clockwise)",
                icon="rotate_right",
                apply_func=partial(_add_spiral, fid_dir=-1),
                apply_for="notes",
            ).props("rounded")
            action_btn_cls(
                tooltip="Random nodes",
                icon="casino",
                apply_func=partial(_add_spiral, fid_dir=0),
                apply_for="notes",
            ).props("rounded outline")
        with ui.row():
            with ui.label("Spikes"):
                wiki_reference("Rail-Options#spikes")
            spike_duration = make_input("Duration", 0, "spike_duration", suffix="b", tooltip="Duration of spikes.")
        with ui.row():
            def _add_spikes(nodes: "numpy array (n, 3)", fid_dir: int, direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,) -> "numpy array (n, 3)":
                if spiral_do_interpolate.value:
                    nodes = rails.interpolate_nodes(nodes, mode="spline", interval=spiral_interpolation.parsed_value)
                return pattern_generation.add_spikes(
                    nodes,
                    fidelity=fid_dir*(360*_safe_inverse(spiral_angle.parsed_value) or 1),  # if angle=0, default to 1
                    radius=spiral_radius.parsed_value,
                    spike_duration=spike_duration.parsed_value,
                    start_angle=spiral_start.parsed_value,
                    direction=direction,
                )
            action_btn_cls(
                tooltip="Spikes (counter-clockwise)",
                icon="rotate_left",
                apply_func=partial(_add_spikes, fid_dir=1),
                apply_for="notes",
            ).props("square")
            action_btn_cls(
                tooltip="Spikes (clockwise)",
                icon="rotate_right",
                apply_func=partial(_add_spikes, fid_dir=-1),
                apply_for="notes",
            ).props("square")
            action_btn_cls(
                tooltip="Spikes (random)",
                icon="casino",
                apply_func=partial(_add_spikes, fid_dir=0),
                apply_for="notes",
            ).props("square")

def wall_spacing_card(action_btn_cls: Any) -> None:
    with ui.card():
        ui.label("Wall spacing")
        with ui.row():
            compress_interval = make_input("Spacing", "1/64", "compress_interval", suffix="b", tooltip="Space between walls")
            action_btn_cls(
                tooltip="Compress (ignores filters)",
                icon="compress",
                icon_angle=90,
                func=lambda data, **kwargs: _space_walls(data, interval=compress_interval.parsed_value),
            )
        with ui.row():
            wall_limit = make_input("Walls/4s", 195, "spawn_limit", tooltip="200=wireframe limit, 500=spawn limit")
            action_btn_cls(
                tooltip="Distribute walls to configured densit (ignores filters)",
                icon="expand",
                icon_angle=90,
                func=lambda data, **kwargs: _space_walls(data, interval=(4*data.bpm/60)/wall_limit.parsed_value),
            )

card_funcs: list[Callable[[Any], None]] = [
    offset_card,
    scaling_card,
    flatten_mirror_card,
    rotate_outset_card,
    rails_card,
    smoothing_card,
    color_card,
    spiral_spike_card,
    wall_spacing_card,
]

def _dashboard_tab() -> None:
    with ui.row():
        with ui.card().classes("h-14 bg-grey-9").props("dark"), ui.row():
            sw_use_orig = ui.switch("", value=False).props('dense size="xl" icon="restore" color="accent" keep-color').bind_value(app.storage.user, "dashboard_use_orig")
            sw_use_orig.tooltip("Restore original data copied from editor, to quickly try different settings.")
            sw_mirror_left = ui.switch("", value=False).props('dense size="xl" icon="flip" color="secondary" keep-color').bind_value(app.storage.user, "dashboard_mirror_left")
            sw_mirror_left.tooltip("Mirror operations for left notes and left walls, e.g. offseting right will move those left instead")
            sw_realign = ui.switch("", value=True).props('dense size="xl" icon="align_horizontal_left" color="orange" keep-color').bind_value(app.storage.user, "dashboard_realign")
            sw_realign.tooltip("After completion, align the start of the selection to the very first object. Not recommended when shifting in time or shortening rails.")

            ui.separator().props("vertical")

            with ui.button(icon="filter_alt", color="info").props("dense").classes("w-10 h-8 -my-1") as filter_menu_btn:
                ui.tooltip("Open filter settings")
                filter_types: dict[str, bool] = app.storage.user.get("dashboard_filter", {ty: True for ty in synth_format.ALL_TYPES})

                filter_badge = ui.badge(str(sum(filter_types.values())), color="green").props("floating")
                def _bdg_color(invert_filter: bool) -> None:
                    filter_badge.props(f'color="{"red" if invert_filter else "green"}"')

                filter_chk: dict[str, ui.checkbox] = {}
                filter_cats: dict[str, ui.checkbox] = {}
                ty_lookup = {"notes": synth_format.NOTE_TYPES, "walls": synth_format.WALL_TYPES, "effects": ("lights", "effects")}

                _chk_update = False
                @contextmanager
                def _chk_change():
                    nonlocal _chk_update
                    if _chk_update:
                        yield True
                    else:
                        _chk_update = True
                        yield False
                        _chk_update = False
                    filter_badge.set_text(str(sum(filter_types.values())))
                    app.storage.user["dashboard_filter"] = filter_types

                def _cat_change(ev: events.ValueChangeEventArguments) -> None:
                    with _chk_change() as recurse:
                        if recurse:
                            return
                        for g, chk in filter_cats.items():
                            if chk != ev.sender:
                                continue
                            for ty in ty_lookup[g]:
                                filter_chk[ty].set_value(ev.value)

                def _ty_change(group: str) -> None:
                    with _chk_change() as recurse:
                        if recurse:
                            return
                        vals = [filter_chk[ty].value for ty in ty_lookup[group]]
                        if all(val == True for val in vals):
                            filter_cats[group].set_value(True)
                        elif all(val == False for val in vals):
                            filter_cats[group].set_value(False)
                        else:
                            filter_cats[group].set_value(None)

                with ui.menu().classes("p-2") as filter_menu:
                    filter_enable = ui.switch("Enable Filter", value=False).props('dense icon="filter_alt" color="info"').classes("p-2").tooltip("Enable filter").bind_value(app.storage.user, "dashboard_filter_enable")
                    filter_menu_btn.bind_icon_from(filter_enable, "value", backward=lambda v: "filter_alt" if v else "filter_alt_off")
                    ui.separator()
                    with ui.element().bind_visibility_from(filter_enable, "value"):
                        with ui.row().classes("bg-grey-9").props("dark"):
                            filter_enable_type = ui.switch("Filter type", value=True).props('dense icon="filter_alt" color="info"').classes("p-2").tooltip("Enable type filter").bind_value(app.storage.user, "dashboard_filter_enable_type")
                            filter_invert = ui.switch("Invert type", value=False).props('dense icon="flaky" color="black"').classes("p-2").tooltip("Invert type filter (ignore selected type)").bind_value(app.storage.user, "dashboard_filter_invert")
                            filter_invert.on_value_change(lambda ev: _bdg_color(ev.value))
                            _bdg_color(filter_invert.value)
                        with ui.element().bind_visibility_from(filter_enable_type, "value"):
                            with ui.row().classes("py-2"):
                                filter_cats["notes"] = ui.checkbox(text="Notes", value=True).props('dense keep-color').classes("w-16").tooltip("Toggle all notes")
                                ui.separator().props("vertical").classes("m-0 p-0")
                                filter_chk["left"] = ui.checkbox(value=True).props('dense color="cyan" keep-color').tooltip("Left notes/rails")
                                filter_chk["right"] = ui.checkbox(value=True).props('dense color="pink" keep-color').tooltip("Right notes/rails")
                                filter_chk["single"] = ui.checkbox(value=True).props('dense color="green" keep-color').tooltip("Single handed notes/rails")
                                filter_chk["both"] = ui.checkbox(value=True).props('dense color="amber" keep-color').tooltip("Two-handed notes/rails")
                            with ui.row().classes("py-2"):
                                filter_cats["walls"] = ui.checkbox(text="Walls", value=True).props('dense keep-color').classes("w-16").tooltip("Toggle all walls")
                                ui.separator().props("vertical").classes("m-0 p-0")
                                with ui.grid(rows=2, columns=4):
                                    filter_chk["wall_left"] = ui.checkbox(value=True).props('dense checked-icon="flip" unchecked-icon="flip" color="cyan"').tooltip("Left walls")
                                    filter_chk["wall_right"] = ui.checkbox(value=True).props('dense checked-icon="flip" unchecked-icon="flip" color="pink"').classes("rotate-180 w-5").tooltip("Right walls")
                                    filter_chk["crouch"] = ui.checkbox(value=True).props('dense checked-icon="flip" unchecked-icon="flip" color="green"').classes("rotate-90 w-5").tooltip("Crouch walls")
                                    filter_chk["square"] = ui.checkbox(value=True).props('dense checked-icon="check_box_outline_blank" unchecked-icon="check_box_outline_blank" color="amber"').tooltip("Square walls")
                                    
                                    filter_chk["angle_left"] = ui.checkbox(value=True).props('dense checked-icon="switch_right" unchecked-icon="switch_right" color="cyan"').tooltip("Left angled walls")
                                    filter_chk["angle_right"] = ui.checkbox(value=True).props('dense checked-icon="switch_left" unchecked-icon="switch_left" color="pink"').tooltip("Right angled walls")
                                    filter_chk["center"] = ui.checkbox(value=True).props('dense checked-icon="unfold_more_double" unchecked-icon="unfold_more_double" color="green"').tooltip("Center walls")
                                    filter_chk["triangle"] = ui.checkbox(value=True).props('dense checked-icon="change_history" unchecked-icon="change_history" color="amber"').tooltip("Triangle walls")
                            with ui.row().classes("py-2"):
                                filter_cats["effects"] = ui.checkbox(text="Effects", value=True).props('dense keep-color').classes("w-16").tooltip("Toggle all effects")
                                ui.separator().props("vertical").classes("m-0 p-0")
                                filter_chk["lights"] = ui.checkbox(value=True).props('dense checked-icon="lightbulb" unchecked-icon="lightbulb" color="yellow"').tooltip("Light effects")
                                filter_chk["effects"] = ui.checkbox(value=True).props('dense checked-icon="bolt" unchecked-icon="bolt" color="yellow"').tooltip("Flash effects")
                            for g, tys in ty_lookup.items():
                                filter_cats[g].on_value_change(_cat_change)
                                for t in tys:
                                    chk = filter_chk[t]
                                    chk.bind_value(filter_types, t)
                                    chk.on_value_change(partial(_ty_change, g))
                                _ty_change(g)  # update group checkbox state
                        ui.separator()
                        with ui.row().classes("bg-grey-9").props("dark"):
                            filter_enable_rails = ui.switch("Filter notes/rails", value=True).props('dense icon="filter_alt" color="info"').classes("p-2").tooltip("Enable notes/rails filter").bind_value(app.storage.user, "dashboard_filter_enable_rail")
                        with ui.element().bind_visibility_from(filter_enable_rails, "value"):
                            with ui.row().classes("bg-grey-9").props("dark"):
                                filter_single = ui.switch("Single notes", value=True).props('dense icon="sports_baseball" color="info"').classes("p-2").tooltip("Allow single notes. Note that some functions need single nodes to work (e.g. split rails).").bind_value(app.storage.user, "dashboard_filter_allow_single")
                                filter_rails = ui.switch("Rails", value=True).props('dense icon="show_chart" color="info"').classes("p-2").tooltip("Allow rails. Length and node spacing can be constrainted below.").bind_value(app.storage.user, "dashboard_filter_allow_rails")
                            with ui.element().bind_visibility_from(filter_rails, "value"):
                                with ui.row():
                                    ui.label("Rail length").classes("my-auto w-24")
                                    ui.separator().props("vertical")
                                    filter_raillen_min = make_input("Min", "", "raillen_min", suffix="b", allow_empty=True)
                                    ui.label("to").classes("my-auto")
                                    filter_raillen_max = make_input("Max", "", "raillen_max", suffix="b", allow_empty=True)
                                    ui.tooltip("Only affect rails with this length").props('max-width="20%"')
                                with ui.row():
                                    ui.label("Node spacing").classes("my-auto w-24")
                                    ui.separator().props("vertical")
                                    filter_railspace_min = make_input("Min", "", "railspace_min", suffix="b", allow_empty=True)
                                    ui.label("to").classes("my-auto")
                                    filter_railspace_max = make_input("Max", "", "railspace_max", suffix="b", allow_empty=True)
                                    ui.tooltip("Only affect rails where all nodes have a time spacing between min and max")
                                with ui.row():
                                    ui.label("Node count").classes("my-auto w-24")
                                    ui.separator().props("vertical")
                                    filter_railnodes_min = make_input("Min", "", "railnodes_min", allow_empty=True)
                                    ui.label("to").classes("my-auto")
                                    filter_railnodes_max = make_input("Max", "", "railnodes_max", allow_empty=True)
                                    ui.tooltip("Only affect rails with a node count between min and max")
                                def _fix_inputs() -> None:
                                    for inp in (filter_raillen_min, filter_raillen_max, filter_railspace_min, filter_railspace_max, filter_railnodes_min, filter_railnodes_max):
                                        inp.update()
                                filter_menu.on_value_change(_fix_inputs)
            ui.separator().props("vertical")

            with ui.label("Coordinates"):
                wiki_reference("Movement-Options#pivot-and-relative")
            coordinate_mode = ui.radio({"absolute":"", "relative":"", "pivot":""}, value="absolute").props("dense inline dark").bind_value(app.storage.user, "dashboard_coord_mode")
            with ui.teleport(f"#c{coordinate_mode.id} > div:nth-child(1)"):
                ui.tooltip("Absolute (grid center)")
                ui.icon("grid_4x4", size="sm")
            with ui.teleport(f"#c{coordinate_mode.id} > div:nth-child(2)"):
                ui.tooltip("Relative (rail head & wall angle)")
                ui.icon("start", size="sm").classes("rotate-315")
            with ui.teleport(f"#c{coordinate_mode.id} > div:nth-child(3)"):
                ui.tooltip("Pivot (custom position)")
                ui.icon("tune", size="sm")
            with ui.row().classes("-my-2") as pivot_settings:
                pivot_settings.bind_visibility_from(coordinate_mode, "value", backward=lambda v: v=="pivot")
                with ui.row():
                    ui.tooltip("Pivot position")
                    pivot_x = make_input("X", 0, "pivot_x", suffix="sq").props("dark")
                    pivot_y = make_input("Y", 0, "pivot_y", suffix="sq").props("dark")
                    pivot_t = make_input("Time", 0, "pivot_t", suffix="b").props("dark")

    class ActionButton(ui.button):
        def __init__(self,
            tooltip: str, wiki_ref: Optional[str] = None,
            icon: str="play_arrow", icon_angle: int=0, 
            func: Callable|None=None,
            apply_func: Callable|None=None,
            apply_for: Literal["all", "walls", "notes"] = "all",
            apply_args: Callable[[], dict[str, ...]]|dict[str, ...]|None=None,
            **kwargs
        ):
            super().__init__(icon=icon if not icon_angle else None, on_click=self.do_action, **kwargs)
            self._tooltip = tooltip
            if (func is not None) == (apply_func is not None):
                raise TypeError("Specify either func OR apply_func")
            if func is not None:
                self._func = func
            else: # apply_func
                if apply_args is None:
                    apply_args = {}
                self._func = lambda data, **kwargs: getattr(data, f"apply_for_{apply_for}")(apply_func, **kwargs, **(apply_args() if callable(apply_args) else apply_args))

            self.classes("w-12 h-10")
            with (self.add_slot("default") if icon_angle else self):
                ui.tooltip(tooltip)
                if wiki_ref is not None:
                    wiki_reference(wiki_ref, True).props("floating")
                if icon_angle:
                    # create dedicated object, which can rotate independently from button
                    ui.icon(icon).style(f"rotate: {icon_angle}deg")


        @handle_errors
        def do_action(self):
            settings = {
                k.removeprefix("dashboard_"): v
                for k, v in app.storage.user.items()
                if k.startswith("dashboard_")
            }
            try:
                rail_filter = None
                types = synth_format.ALL_TYPES
                if filter_enable.value and filter_enable_rails.value:
                    rail_filter = synth_format.RailFilter(
                        single=filter_single.value,
                        rails=filter_rails.value,
                        min_len=filter_raillen_min.parsed_value,
                        max_len=filter_raillen_max.parsed_value,
                        min_count=filter_railnodes_min.parsed_value,
                        max_count=filter_railnodes_max.parsed_value,
                        min_spacing=filter_railspace_min.parsed_value,
                        max_spacing=filter_railspace_max.parsed_value,
                    )
                if filter_enable.value and filter_enable_type.value:
                    types = tuple(ty for ty, ty_enabled in filter_types.items() if ty_enabled != filter_invert.value)
                with safe_clipboard_data(use_original=sw_use_orig.value, realign_start=sw_realign.value) as data:
                    self._func(
                        data=data,
                        types=types,
                        rail_filter=rail_filter or None,  # if rail-filter is false-ish (nothing set), pass None
                        mirror_left=sw_mirror_left.value,
                        relative=(coordinate_mode.value == "relative"),
                        pivot=np.array([pivot_x.parsed_value, pivot_y.parsed_value, pivot_t.parsed_value]) if (coordinate_mode.value=="pivot") else None
                    )
            except (PrettyError, ParseInputError):
                raise
            except Exception as exc:
                raise PrettyError(
                    msg=f"Error executing '{self._tooltip}'",
                    exc=exc,
                    context={"settings": settings},
                ) from exc
            counts = data.get_counts()
            info(
                f"Completed: '{self._tooltip}'",
                caption=pretty_list(
                    [f"{counts[t]['total']} {t if counts[t]['total'] != 1 else t.rstrip('s')}" for t in ("notes", "rails", "rail_nodes", "walls")]
                    +([f"{len(types)} types filtered"]  if set(types) != set(synth_format.ALL_TYPES) else [])
                    +(["note/rail filter active"] if rail_filter else [])
                ),
            )

    with pivot_settings:
        def _pick_pivot(data: synth_format.DataContainer, types, rail_filter, **kwargs) -> None:
            first = data.find_first()
            if first is None:
                raise PrettyError("No object found matching the filter")
            x, y, t = first[1][:3]
            pivot_x.value = pretty_fraction(x)
            pivot_y.value = pretty_fraction(y)
            pivot_t.value = pretty_fraction(t)
        ActionButton(
            tooltip="Pick position of first object",
            icon="colorize",
            func=_pick_pivot,
        ).props("outline").classes(replace="w-10 h-8 my-1")

    with ui.row():
        for card_func in card_funcs:
            card_func(ActionButton)

dashboard_tab = GUITab(
    name = "dashboard",
    label = "Dashboard",
    icon = "dashboard",
    content_func=_dashboard_tab,
)


import dataclasses
from io import BytesIO

import librosa
import numpy as np
import soundfile


OGG_WRITE_CHUNK_SIZE = 0x10000

class AudioNotOggError(RuntimeError):
    def __init__(self, detected_format: str) -> None:
        self.detected_format = detected_format

@dataclasses.dataclass
class AudioData:
    raw_data: bytes
    sample_rate: int
    channels: int
    duration: float

    @staticmethod
    def from_raw(raw_data: bytes, *, allow_conversion: bool = False) -> "AudioData":
        try:
            info = soundfile.info(BytesIO(raw_data))
        except soundfile.SoundFileError as sfe:
            raise ValueError(f"Could not parse audio file: {sfe!r}")
        if info.format != "OGG":
            if not allow_conversion:
                raise AudioNotOggError(info.format)
            data, sr = soundfile.read(BytesIO(raw_data))
            raw_data = export_ogg(data.T, samplerate=sr)
        return AudioData(
            raw_data=raw_data,
            sample_rate=info.samplerate,
            channels=info.channels,
            duration=info.duration,
        )

    def with_silence(self, before_start_s: float = 0, after_end_s: float = 0) -> "AudioData":
        if not before_start_s and not after_end_s:
            return self
        data, sr = soundfile.read(BytesIO(self.raw_data))
        if before_start_s < 0:
            data = data[librosa.time_to_samples(-before_start_s, sr=sr):]
        if before_start_s > 0:
            p = np.zeros_like(data, shape=(librosa.time_to_samples(before_start_s, sr=sr), *data.shape[1:]))
            data = np.concatenate((p, data))
        if after_end_s < 0:
            data = data[:-librosa.time_to_samples(-after_end_s, sr=sr)]
        if after_end_s > 0:
            p = np.zeros_like(data, shape=(librosa.time_to_samples(after_end_s, sr=sr), *data.shape[1:]))
            data = np.concatenate((data, p))
        return AudioData(
            raw_data=export_ogg(data.T, samplerate=sr),
            sample_rate=sr,
            channels=data.shape[1],
            duration=librosa.samples_to_time(data.shape[0], sr=sr),
        )

def load_for_analysis(raw_data: bytes) -> tuple["numpy array (f,)", int]:
    data, sr = librosa.load(BytesIO(raw_data))  # load with default samplerate and as mono
    return data, int(sr)

def export_ogg(data: "numpy array (c, s)|(s,)", samplerate: int = 22050) -> bytes:
    # librosa uses channels x samples instead of samples x channels, so transpose
    data = data.T
    bio = BytesIO()
    # saving as ogg segfauls if the chunks are too big, so chunk the writes
    # based on https://github.com/bastibe/python-soundfile/issues/426#issuecomment-2150934383
    if data.ndim == 1:
        channels = 1
    else:
        channels = data.shape[1]
    with soundfile.SoundFile(bio, 'w', samplerate=samplerate, channels=channels, format="ogg") as f:
        num_chunks = (len(data) + OGG_WRITE_CHUNK_SIZE - 1) // OGG_WRITE_CHUNK_SIZE
        for chunk in np.array_split(data, num_chunks, axis=0):
            f.write(chunk)

    return bio.getvalue()

def export_wav(data: "numpy array (c, s)|(s,)", samplerate: int = 22050) -> bytes:
    # librosa uses channels x samples instead of samples x channels, so transpose
    data = data.T
    bio = BytesIO()
    soundfile.write(file=bio, data=data, samplerate=samplerate, format="wav")
    return bio.getvalue()

def audio_with_clicks(raw_audio_data: bytes, duration: float, bpm: float, offset_ms: int) -> bytes:
    beat_time = 60/bpm
    data, sr = librosa.load(BytesIO(raw_audio_data))
    clicks = librosa.clicks(times=np.arange(beat_time-(offset_ms/1000)%beat_time, duration, beat_time), length=len(data), sr=sr)
    return export_ogg(data+clicks, samplerate=int(sr))

def find_trims(raw_audio_data: bytes) -> tuple[float, float]:
    data, sr = librosa.load(BytesIO(raw_audio_data))
    _, (start, end) = librosa.effects.trim(data)
    return librosa.samples_to_time(start), librosa.samples_to_time(data.shape[0]-end)


# based on: https://github.com/zauberzeug/nicegui/blob/main/examples/local_file_picker/local_file_picker.py
import platform
from pathlib import Path
from typing import Optional

from nicegui import events, ui

class local_dir_picker(ui.dialog):

    def __init__(self, directory: str) -> None:
        """Local File Picker

        This is a simple file picker that allows you to select a file from the local filesystem where NiceGUI is running.

        :param directory: The directory to start in.
        :param upper_limit: The directory to stop at (None: no limit, default: same as the starting directory).
        """
        super().__init__()
        with self, ui.card():
            self.add_drives_toggle()
            self.path_input = ui.input(
                "Directory",
                on_change=lambda e: self.set_path(e.value),
                validation={"Not a directory": lambda v: v and Path(v).is_dir()}
            ).props("no-error-icon").classes("w-full")
            self.grid = ui.aggrid({
                'columnDefs': [{'field': 'name', 'headerName': 'Directory'}],
            }, html_columns=[0]).classes('w-96').on('cellClicked', lambda e: self.set_path(e.args['data']['path']))
            with ui.row().classes('w-full justify-end'):
                with ui.button('Cancel', on_click=self.close).props('outline'):
                    ui.tooltip(f"Keeps current path: {directory}")
                ui.button('Ok', on_click=lambda v: self.submit(self.path))
        self.set_path(directory)


    def add_drives_toggle(self):
        with ui.row():
            with ui.button(icon="home", on_click=lambda _: self.set_path("~")):
                ui.tooltip(f"User home: {Path('~').expanduser()}")
            with ui.button(icon="terminal", on_click=lambda _: self.set_path(str(Path().absolute()))):
                ui.tooltip(f"Current directory: {Path().absolute()}")
            if platform.system() == 'Windows':
                import win32api
                drives = win32api.GetLogicalDriveStrings().split('\000')[:-1]
                ui.toggle(drives, value=drives[0], on_change=lambda e: self.set_path(e.value))
            else:
                with ui.button("/", on_click=lambda _: self.set_path("/")):
                    ui.tooltip("Filesystem root")

    def set_path(self, p: str):
        pp = Path(p).expanduser().absolute()
        if p and pp.is_dir():
            self.path = pp
            self.path_input.value = str(pp)
            self.path_input.set_autocomplete(list(str(p) for p in pp.iterdir() if p.is_dir()))
            self._update_grid()

    def _update_grid(self) -> None:
        paths = list(p for p in self.path.iterdir() if p.is_dir())
        paths.sort(key=lambda p: p.name.lower())

        self.grid.options['rowData'] = [
            {
                'name': f'📁 <strong>{p.name}</strong>',
                'path': str(p),
            }
            for p in paths
        ]
        if self.path != self.path.parent:
            self.grid.options['rowData'].insert(0, {
                'name': '↖️ <strong>..</strong>',
                'path': str(self.path.parent),
            })
        self.grid.update()



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from io import BytesIO
import json
from json import JSONDecodeError
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.ticker import FuncFormatter

from . import synth_format, __version__
from .rails import interpolate_spline


min_sec_formatter = FuncFormatter(lambda x,_: f"{int(x//60):d}:{int(x%60):02d}{'.{:03d}'.format(int((x%1)*1000)) if x%1 else ''}")

FINALIZED_BOOKMARK = {"time": 0.0, "name": "#smh_finalized"}
BEATMAP_JSON_FILE = "beatmap.meta.bin"
DIFFICULTIES = ("Easy", "Normal", "Hard", "Expert", "Master", "Custom")
NOTE_COLORS = {"right": "red", "left": "cyan", "single": "lime", "both": "gold"}

RENDER_WINDOW = 4  # 4 seconds of elements are rendered
WALL_DESPAWN_PC = 80  # limit of visible walls, earlier ones despawn
WALL_DESPAWN_QUEST = 40

def get_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        prog=f"python3 -m {__package__}.{Path(__file__).stem}",
        description='\n'.join([
            "[DEPRECATED] Finalizes a map:"
            "  * Move certain walls up so they appear ingame as they do in the editor"
            "",
            "Also see the wiki on GitHub, which contains more detailed explainations, as well as some examples and images: https://github.com/adosikas/synth_mapping_helper/wiki",
        ]),
        epilog=f"Version: {__version__}",
    )
    parser.add_argument("input", type=Path, help="Input file")
    parser.add_argument("output", type=Path, help="Output file")
    parser.add_argument("--revert", action="store_true", help="Reverse finalizing options")
    parser.add_argument("--plot-notes", action="store_true", help="Plot note velocity and acceleration")
    parser.add_argument("--plot-walls", action="store_true", help="Plot walls for pc and quest")
    return parser

def abort(reason: str):
    print("ERROR: " + reason)
    exit(1)

def main(options):
    if not options.input.is_file():
        abort("Input file is not a file, is the path correct?")
    out_buffer = BytesIO()  # buffer output zip file in memory, only write on success
    with ZipFile(options.input) as inzip, ZipFile(out_buffer, "w") as outzip:
        # copy all content except beatmap json
        outzip.comment = inzip.comment
        for info in inzip.infolist():
            if info.filename != BEATMAP_JSON_FILE:
                outzip.writestr(info, inzip.read(info.filename))
        # load beatmap json
        beatmap = json.loads(inzip.read(BEATMAP_JSON_FILE))
        bpm = beatmap["BPM"]
        finalized = FINALIZED_BOOKMARK in beatmap["Bookmarks"]["BookmarksList"]
        if not options.revert:
            if finalized:
                abort("Already finalized!")
        else:
            if not finalized:
                abort("Not finalized, will not revert!")
            beatmap["Bookmarks"]["BookmarksList"].remove(FINALIZED_BOOKMARK)
            
        # shift slides up so they appear ingame as they do in the editor
        for _, walls in beatmap["Slides"].items():
            for w in walls:
                if options.revert: 
                    w["position"][1] -= 2.1 * synth_format.GRID_SCALE
                else:
                    w["position"][1] += 2.1 * synth_format.GRID_SCALE

        for diff in DIFFICULTIES:
            # skip difficulties without notes
            if not beatmap["Track"][diff]:
                continue
            if options.plot_notes:
                fig, axs = plt.subplots(4, 1, sharex=True, figsize=(16, 8))
                fig.suptitle(f"{beatmap['Author']} - {beatmap['Name']}: {diff} / Notes")
                (ax_x, ax_y, ax_vel, ax_acc) = axs
                ax_x.set_ylabel("X-Position (sq)")
                ax_x.set_ylim((8, -8))
                ax_x.grid(True)
                ax_y.set_ylabel("Y-Position (sq)")
                ax_y.set_ylim((-6, 6))
                ax_y.grid(True)

                ax_vel.set_ylabel("Velocity (sq/s)")
                ax_vel.set_ylim((0, 50))
                ax_acc.set_ylabel("Acceleration (sq/s²)")
                ax_acc.set_ylim((0, 50))

                axs[-1].set_xlabel("time (s)")
                axs[-1].xaxis.set_major_formatter(min_sec_formatter)
                # bookmarks
                for bookmark in beatmap["Bookmarks"]["BookmarksList"]:
                    time = utils.beat_to_second(int(bookmark["time"]) / 64, bpm)
                    for ax in axs:
                        ax.axvline(time, color="grey")
                    axs[0].text(time, 0.99, bookmark["name"], ha='left', va='bottom', rotation=45, transform=axs[0].get_xaxis_transform())
                
                # notes & rails
                positions = [{} for _ in synth_format.NOTE_TYPES]
                notes_dict = beatmap["Track"][diff]

                for time in sorted(notes_dict):
                    note_list = notes_dict[time]
                    for note in note_list:
                        note_type, pos = synth_format.note_from_synth(bpm, 0, note)
                        color = NOTE_COLORS[synth_format.NOTE_TYPES[note_type]]
                        if pos.shape[0] != 1:
                            new_times = np.arange(pos[0,2], pos[-1,2], 1/64)
                            pos = interpolate_spline(pos, new_times)
                        pos[:, 2] *= 60 / bpm  # convert beat to second
                        for p in pos:
                            positions[note_type][p[2]] = p[:2]

                        # single note / rail head
                        ax_x.plot(pos[0, 2], pos[0, 0], color=color, marker="o")
                        ax_y.plot(pos[0, 2], pos[0, 1], color=color, marker="o")
                        if pos.shape[0] != 1:
                            # rail
                            ax_x.plot(pos[:, 2], pos[:, 0], color=color, marker="")
                            ax_y.plot(pos[:, 2], pos[:, 1], color=color, marker="")

                for note_type, poss in enumerate(positions):
                    if len(poss) <= 1:
                        continue
                    times = sorted(poss)
                    prev_t = times[0]
                    vels = []
                    accs = []
                    for t in times[1:]:
                        if t - prev_t < 1:
                            vel = (poss[t] - poss[prev_t])/(t-prev_t)
                            if vels and vels[-1][1] is not None:
                                accs.append((prev_t, vel - vels[-1][1]))
                            else:
                                accs.append((prev_t, None))
                            vels.append(((t+prev_t)/2, vel))
                        else:
                            vels.append(((t+prev_t)/2, None))
                        prev_t = t
                    ax_vel.plot([t for t, _ in vels], [np.sqrt(v.dot(v)) if v is not None else np.nan for _, v in vels], color=NOTE_COLORS[synth_format.NOTE_TYPES[note_type]])
                    ax_acc.plot([t for t, _ in accs], [np.sqrt(a.dot(a)) if a is not None else np.nan for _, a in accs], color=NOTE_COLORS[synth_format.NOTE_TYPES[note_type]])

                fig.tight_layout()
                plt.show()
                # fig.savefig(f"{diff}_notes.png")

            if options.plot_walls:
                fig, axs = plt.subplots(4, 1, sharex=True, figsize=(16, 8))
                fig.suptitle(f"{beatmap['Author']} - {beatmap['Name']}: {diff} / Walls")
                (ax_density_pc, ax_pc, ax_density_quest, ax_quest) = axs
                ax_density_pc.set_title("PC")
                ax_density_pc.set_ylabel("Walls visible")
                ax_density_pc.set_ylim([0, 120])
                ax_density_pc.axhline(80, color="red")
                ax_pc.set_ylabel("Wall Status")
                ax_pc.set_yticks([])
                ax_pc.set_ylim((0,11))
            
                ax_density_quest.set_title("Quest")
                ax_density_quest.set_ylabel("Walls visible")
                ax_density_quest.set_ylim([0, 60])
                ax_density_quest.axhline(40, color="red")
                ax_quest.set_ylabel("Wall Status")
                ax_quest.set_yticks([])
                ax_quest.set_ylim((0,11))

                axs[-1].set_xlabel("time (s)")
                axs[-1].xaxis.set_major_formatter(min_sec_formatter)

                # bookmarks
                for bookmark in beatmap["Bookmarks"]["BookmarksList"]:
                    time = utils.beat_to_second(int(bookmark["time"]) / 64, bpm)
                    for ax in axs:
                        ax.axvline(time, color="grey")
                    axs[0].text(time, 0.99, bookmark["name"], ha='left', va='bottom', rotation=45, transform=axs[0].get_xaxis_transform())


                # walls
                all_walls: dict[str, list[float]] = {w_type: [] for w_type in synth_format.WALL_TYPES}
                for slide in beatmap["Slides"][diff]:
                    all_walls[synth_format.WALL_LOOKUP[slide["slideType"]]].append(utils.beat_to_second(slide["time"] / 64, bpm))
                for k, walls in all_walls.items():
                    all_walls[k] = sorted(walls)
                all_walls["crouch"] = sorted(utils.beat_to_second(w["time"] / 64, bpm) for w in beatmap["Crouchs"][diff])
                all_walls["square"] = sorted(utils.beat_to_second(w["time"] / 64, bpm) for w in beatmap["Squares"][diff])
                all_walls["triangle"] = sorted(utils.beat_to_second(w["time"] / 64, bpm) for w in beatmap["Triangles"][diff])

                wall_markers = {
                    "wall_left": ("s", "left", 7),
                    "wall_right": ("s", "right", 6),
                    "angle_left": ("o", "left", 5),
                    "angle_right":  ("o", "right", 4),
                    "center": ("d", "full", 3),
                    "crouch": ("s", "top", 2),
                    "triangle": ("^", "none", 1),
                    "square": ("s", "none", 0),
                }
                # ceil wall delay to next 1/64
                seconds_per_tick = utils.beat_to_second(1/64)
                quest_wall_delay = np.ceil(0.05 / seconds_per_tick) * seconds_per_tick

                for wall_type, (marker, fill, y) in wall_markers.items():
                    density_pc: list[tuple[float, int]] = []
                    density_quest: list[tuple[float, int]] = []

                    pc_ok: list[float] = []
                    pc_despawn: list[float] = []
                    quest_hidden: list[float] = []
                    quest_nothidden: list[float] = []
                    quest_ok: list[float] = []
                    quest_despawn: list[float] = []

                    last_time_quest = None
                    for i, time in enumerate(all_walls[wall_type]):
                        # pc - filter out despawn
                        visible_count = 1
                        end = time + RENDER_WINDOW
                        for other_time in all_walls[wall_type][i+1:]:
                            if other_time > end:
                                break
                            visible_count += 1
                        density_pc.append((time, visible_count))
                        if visible_count > WALL_DESPAWN_PC:
                            pc_despawn.append(time)
                        else:
                            pc_ok.append(time)
                        # quest - first step: filter out hidden ones
                        hidden_on_quest = last_time_quest is not None and time - last_time_quest < quest_wall_delay
                        if hidden_on_quest:
                            quest_hidden.append(time)
                        else:
                            last_time_quest = time
                            quest_nothidden.append(time)
        
                    # quest - second step: filter out despawn
                    for i, time in enumerate(quest_nothidden):
                        visible_count = 1
                        end = time + RENDER_WINDOW
                        for other_time in quest_nothidden[i+1:]:
                            if other_time > end:
                                break
                            visible_count += 1
                        density_quest.append((time, visible_count))
                        if visible_count > WALL_DESPAWN_QUEST:
                            quest_despawn.append(time)
                        else:
                            quest_ok.append(time)

                    

                    # pc: ok first, despawn on top
                    for walls, color in zip((pc_ok, pc_despawn), ("green", "orange")):
                        for time in walls:
                            ax_pc.plot([time], [y], marker=marker, fillstyle=fill, color=color)
                    ax_density_pc.plot([time for time, _ in density_pc], [count for _, count in density_pc], label=wall_type)
        
                    # quest: ok first, then despawn, hidden on top
                    for walls, color in zip((quest_ok, quest_despawn, quest_hidden), ("green", "orange", "red")):
                        for time in walls:
                            ax_quest.plot([time], [y], marker=marker, fillstyle=fill, color=color)
                    ax_density_quest.plot([time for time, _ in density_quest], [count for _, count in density_quest], label=wall_type)


                legend_elements = [
                    Line2D([0], [0], color='green', label='ok', marker="o", linestyle=""),
                    Line2D([0], [0], color='orange', label='despawn', marker="o", linestyle=""),
                    Line2D([0], [0], color='red', label='hidden', marker="o", linestyle=""),
                ]
                ax_density_pc.legend(loc="upper right", ncol=len(wall_markers))
                ax_pc.legend(handles=legend_elements[:2], loc="upper right", ncol=2)
                ax_density_quest.legend(loc="upper right", ncol=len(wall_markers))
                ax_quest.legend(handles=legend_elements, loc="upper right", ncol=3)

                fig.tight_layout()
                plt.show()
                # fig.savefig(f"{diff}_walls.png")

        if not options.revert:
            beatmap["Bookmarks"]["BookmarksList"].append(FINALIZED_BOOKMARK)
        # write modified beatmap json
        outzip.writestr(inzip.getinfo(BEATMAP_JSON_FILE), json.dumps(beatmap))
    # write output zip
    options.output.write_bytes(out_buffer.getbuffer())

if __name__ == "__main__":
    main(get_parser().parse_args())



from multiprocessing import freeze_support
freeze_support()

from argparse import ArgumentParser
from pathlib import Path
import logging
import sys

from nicegui import app, ui
import requests

from synth_mapping_helper import __version__
from synth_mapping_helper.gui_tabs.utils import *
from synth_mapping_helper.gui_tabs.commands import command_tab
from synth_mapping_helper.gui_tabs.file_utils import file_utils_tab
from synth_mapping_helper.gui_tabs.dashboard import dashboard_tab
from synth_mapping_helper.gui_tabs.autobackup import autobackup_tab
from synth_mapping_helper.gui_tabs.version import version_tab
from synth_mapping_helper.gui_tabs.stacking import stacking_tab
from synth_mapping_helper.gui_tabs.text_gen import text_gen_tab
from synth_mapping_helper.gui_tabs.wall_art import wall_art_tab

tab_list: list[GUITab] = [
    dashboard_tab,
    stacking_tab,
    text_gen_tab,
    wall_art_tab,
    command_tab,
    file_utils_tab,
    autobackup_tab,
    version_tab,
]

version = f"SMH-GUI v{__version__}"

@app.get("/version")
def get_version():
    return version

async def stop():
    logger.info("Stopping...")
    await ui.run_javascript("setTimeout(window.close, 100);")
    app.shutdown()

def entrypoint():
    parser = ArgumentParser(description=version)
    parser.add_argument("-l", "--log-level", type=str, help="Set log level")
    parser.add_argument("--host", type=str, default="127.0.0.1",
        help="""Host for the webserver. Defaults to 127.0.0.1 (localhost).
            Can be set to a local IP if you want to access the GUI from another device, eg. tablet.\n
            Note that there is NO PASSWORD CHECK, so only use if you trust ALL devices on that network to have access to your clipboard and files.""")
    parser.add_argument("--port", type=int, default=8080, help="Port for the webserver")
    parser.add_argument("--background", action="store_true", help="Open in background (does not open browser)")
    parser.add_argument("--dev-mode", action="store_true", help="Open in dev mode (reloads when editing python files)")

    args = parser.parse_args()
    if args.log_level is None:
        args.log_level = "DEBUG" if args.dev_mode else "INFO"

    if not args.dev_mode and __name__ == "__main__":  # don't check in dev mode or when spawned as child process
        try:
            resp = requests.get(f"http://{args.host}:{args.port}/version", timeout=1)
            resp.raise_for_status()
            print(f"ERROR: {resp.json()} is already running on http://{args.host}:{args.port}")
            sys.exit(-1)
        except requests.ConnectionError:
            # we want an connection error to occur, else there is another instance (or something else) running
            pass
        except Exception as e:
            # unexpected error getting or parsing version
            print(f"ERROR: Could not check if there is another instance already running on http://{args.host}:{args.port}")
            print(f"           {e!r}")
            print("       If this persists after a restart, something else may be using that port and you could add e.g. --port=8181 to change the SMH-GUI port")
            sys.exit(-1)

    @ui.page("/")
    def index():
        ui.add_head_html("""<style>
            .q-field__suffix {
                color: grey !important;
            }
            .q-field__bottom {
                min-height: 0 !important;
                padding: 0 !important;
            }
            .q-uploader__header-content {
                padding: 0px !important;
            }
            .q-uploader__list {
                min-height: 0 !important;
                padding: 0 !important;
            }
        </style>""")
        with ui.header(elevated=True):
            with ui.tabs() as tabs:
                tab_containers = [
                    ui.tab(name=t.name, label=t.label, icon=t.icon)
                    for t in tab_list
                ]
            with ui.button(icon="bug_report" + "support", color="negative", on_click=lambda _:ui.download.from_url("error_report", filename="smh_gui_error.json")).props("text-color=white").classes("ml-auto"):
                ui.tooltip("Save report of last error")
            with ui.element():
                ui.tooltip("Switch dark mode")
                dark = ui.dark_mode(True)  # start in dark mode (matching editor)
                ui.button(icon="dark_mode", on_click=dark.enable).bind_visibility_from(dark, "value", backward=lambda v: v is not True).props('text-color=white')
                ui.button(icon="light_mode", on_click=dark.disable).bind_visibility_from(dark, "value", backward=lambda v: v is not False).props('text-color=white')

            with ui.button(icon="question_mark", color="white", on_click=lambda _:ui.navigate.to(wiki_base, new_tab=True)).props("text-color=primary"):
                ui.tooltip("Open wiki")
            with ui.button(icon="close", color="red", on_click=stop) as close_button:
                ui.tooltip("Quit")
                
        with ui.tab_panels(tabs, value=tab_list[0].name).classes("w-full").bind_value(app.storage.user, "active_tab"):
            for t in tab_list:
                with ui.tab_panel(t.name) as panel:
                    try:
                        t.content_func()
                    except Exception as exc:
                        panel.clear()
                        error(f"Error loading {t.label} tab", exc=exc, context=t.name, data=t.get_settings())
                        ui.label(f"Error loading {t.label} tab: {exc!r}")
                        ui.label("Consider sending me the error report, so this can be avoided in the future.")
                        ui.label("You may try reverting to the default settings below.")
                        ui.button("Delete settings", icon="delete", color="negative", on_click=t.delete_settings)

    logging.basicConfig(
        format='%(asctime)s %(levelname)-8s [%(name)s] %(message)s',
        level=args.log_level,
        datefmt='%Y-%m-%d %H:%M:%S',
        handlers=[
            logging.FileHandler("smh_gui.log"),
            logging.StreamHandler()
        ]
    )
    # hide some spammy logs
    for ln in ("watchfiles", "multipart", "python_multipart", "numba"):
        logging.getLogger(ln).setLevel(logging.WARN)

    if __name__ == "__main__":  # don't run when spawned as child process
        logger.info(f"Starting {version}{' in background' if args.background else ''}. Working directory: {Path().absolute()}")
    ui.run(
        host=args.host,
        port=args.port,
        title=version,
        favicon="🚧" if args.dev_mode else "🤦",
        reload=args.dev_mode,
        storage_secret="smh_gui",
        show=not args.background,
    )

if __name__ in {"__main__", "__mp_main__"}:
    entrypoint()



from pathlib import Path
import time

from nicegui import app, events, ui
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, FileSystemEvent

from .local_dir_picker import local_dir_picker
from .utils import *

def _isdir(v: str) -> bool:
    return Path(v).is_dir()

last_check: dict[Path, float] = {}
last_backup: dict[Path, tuple[float, Path]] = {}

def _autobackup_tab() -> None:
    watcher_dir: Path|None = None
    def watcher_func() -> None:
        global last_check
        if not watcher_dir or "autobackup_backupdir" not in app.storage.user or "autobackup_minage" not in app.storage.user:
            return
        now = time.time()
        for p in watcher_dir.glob("*.synth"):
            if not p.is_file():
                continue
            if p.stat().st_mtime > last_check.get(watcher_dir, now):
                logger.debug(f"File change detected: {p}")
                if now - last_backup.get(p, (0,None))[0] > app.storage.user["autobackup_minage"]:
                    pb = Path(app.storage.user["autobackup_backupdir"]) / f"{p.stem}.{time.strftime('%Y-%m-%d_%H-%M-%S')}{p.suffix}"
                    logger.info(f"{p} changed, creating backup: {pb}")
                    pb.write_bytes(p.read_bytes())
                    last_backup[p] = (now, pb)
        last_check[watcher_dir] = now

        log.refresh()

    watcher = ui.timer(60, watcher_func, active=False)

    async def pick_dir(e: events.ClickEventArguments) -> None:
        key = "autobackup_workdir" if e.sender == workdir_picker else "autobackup_backupdir"
        result = await local_dir_picker(app.storage.user.get(key) or ".")
        if result is not None:
            input_element: ui.input = e.sender.parent_slot.parent  # type: ignore
            input_element.value = str(result)

    def dir_changed() -> None:
        nonlocal watcher_dir
        if not app.storage.user.get("autobackup_enabled"):
            watcher_dir = None
            watcher.active = False
            return
        p = app.storage.user.get("autobackup_workdir")
        if not p or not Path(p).is_dir():
            watcher_dir = None
            watcher.active = False
            return
        if watcher_dir == Path(p):
            return
        info(f"Now watching {p} for .synth file changes")
        watcher_dir = Path(p)
        last_check[Path(p)] = time.time()
    dir_changed()

    with ui.row():
        ui.switch("Enable Autobackup", value=False, on_change=dir_changed).classes("my-auto").bind_value(watcher, "active").bind_value(app.storage.user, "autobackup_enabled")
        with ui.number("Check interval", value=60, suffix="s", min=15).props('input-style="text-align: right"').classes("w-20").bind_value(watcher, "interval").bind_value(app.storage.user, "autobackup_interval"):
            ui.tooltip("Amount of time between checks")
        with ui.number("Minimum age", value=300, suffix="s").props('input-style="text-align: right"').classes("w-20").bind_value(app.storage.user, "autobackup_minage"):
            ui.tooltip("If > 0, backups are not created if the last backup was less than X seconds ago")
        with ui.input("Editor working directory", validation={"Not a directory": _isdir}, on_change=dir_changed).classes("w-96").bind_value(app.storage.user, "autobackup_workdir"):
            workdir_picker = ui.button(on_click=pick_dir, icon="pageview").props("outline").classes("my-auto")
            ui.tooltip("Directory with the .synth files you want to monitor")
        with ui.input("Backup directory", validation={"Not a directory": _isdir}, on_change=dir_changed).classes("w-96").bind_value(app.storage.user, "autobackup_backupdir"):
            backupdir_picker = ui.button(on_click=pick_dir, icon="archive").props("outline").classes("my-auto")
            ui.tooltip("Directory to save the backups to")

    @ui.refreshable
    def log() -> None:
        for _, b in sorted(last_backup.values(), key=lambda tb: tb[0], reverse=True):
            ui.label(str(b))

    ui.label("Last backup:").classes("my-4")
    with ui.card().classes("w-full"):
            log()

autobackup_tab = GUITab(
    name="autobackup",
    label="AutoBackup",
    icon="manage_history",
    content_func=_autobackup_tab,
)



from dataclasses import dataclass, field

from nicegui import app, elements, events, ui
import numpy as np

from synth_mapping_helper.gui_tabs.utils import SMHInput, handle_errors
from synth_mapping_helper.utils import parse_number, pretty_fraction
from synth_mapping_helper import synth_format

GRID_SIZE = (8, 6)

@dataclass
class ObjectSettings:
    color: str
    opacity: float
    size: float

@dataclass
class RenderSettings:
    color_left: str = "#4ff"
    color_right: str = "#f4f"
    color_single: str = "#4f4"
    color_both: str = "#ff4"

    note: ObjectSettings = field(default_factory=lambda:ObjectSettings("", 1.0, 0.6))
    rail: ObjectSettings = field(default_factory=lambda:ObjectSettings("", 0.5, 0.2))
    rail_node: ObjectSettings = field(default_factory=lambda:ObjectSettings("", 1.0, 0.3))

    wall: ObjectSettings = field(default_factory=lambda:ObjectSettings("#808", 0.1, 1/64))
    wall_outline: ObjectSettings = field(default_factory=lambda:ObjectSettings("#808", 1, 0.0))

DEFAULT_SETTINGS = RenderSettings()

WALL_VERTS = {
    i: np.array(synth_format.WALL_VERTS[synth_format.WALL_LOOKUP[i]])
    for i in synth_format.WALL_LOOKUP
}

def make_input(label: str, value: str|float, storage_id: str, **kwargs) -> SMHInput:
    default_kwargs: dict[str, str|int] = {"tab_id": "render", "width": 14}
    return SMHInput(storage_id=storage_id, label=label, default_value=value, **(default_kwargs|kwargs))

class SettingsPanel(ui.element):
    def __init__(self) -> None:
        super().__init__()
        with self:
            ui.label("Note: These settings only affect the preview.").tooltip("The game does not allow maps to override colors.")
            with ui.row():
                self.wall_size = make_input("Wall Depth", pretty_fraction(DEFAULT_SETTINGS.wall.size), "wall_size", suffix="b")
                ui.separator().props("vertical")
                self.wall_color = ui.color_input("Wall", value=DEFAULT_SETTINGS.wall.color, preview=True).props("dense").classes("w-24").bind_value(app.storage.user, "render_wall_color")
                self.wall_color.button.style("color: black")
                self.wall_opacity = make_input("Opacity", DEFAULT_SETTINGS.wall.opacity, "wall_opacity")
                ui.separator().props("vertical")
                self.wall_outline_color = ui.color_input("Outline", value=DEFAULT_SETTINGS.wall_outline.color, preview=True).props("dense").classes("w-24").bind_value(app.storage.user, "render_wall_outline_color")
                self.wall_outline_color.button.style("color: black")
                self.wall_outline_opacity = make_input("Opacity", DEFAULT_SETTINGS.wall_outline.opacity, "wall_outline_opacity")
            ui.separator().classes("my-2")
            with ui.row():
                self.note_colors: dict[str, ui.color_input] = {
                    t: ui.color_input(t.capitalize(), value=getattr(DEFAULT_SETTINGS, "color_"+t), preview=True).props("dense").classes("w-24 h-10")
                    for t in synth_format.NOTE_TYPES
                }
                for t, ci in self.note_colors.items():
                    ci.bind_value(app.storage.user, "render_color_"+t)
                    ci.button.style("color: black")
            ui.separator().classes("my-2")
            with ui.row():
                self.note_size = make_input("Note Size", DEFAULT_SETTINGS.note.size, "note_size", suffix="sq")
                self.note_opacity = make_input("Opacity", DEFAULT_SETTINGS.note.opacity, "note_opacity")
                ui.separator().props("vertical")
                self.rail_size = make_input("Rail Size", DEFAULT_SETTINGS.rail.size, "rail_size", suffix="sq")
                self.rail_opacity = make_input("Opacity", DEFAULT_SETTINGS.rail.opacity, "rail_opacity")
                ui.separator().props("vertical")
                self.rail_node_size = make_input("Node Size", DEFAULT_SETTINGS.rail_node.size, "rail_node_size", suffix="sq")
                self.rail_node_opacity = make_input("Opacity", DEFAULT_SETTINGS.rail_node.opacity, "rail_node_opacity")

    def parse_settings(self) -> RenderSettings:
        return RenderSettings(
            color_left=self.note_colors["left"].value,
            color_right=self.note_colors["right"].value,
            color_single=self.note_colors["single"].value,
            color_both=self.note_colors["both"].value,
            note=ObjectSettings("", self.note_opacity.parsed_value, self.note_size.parsed_value),
            rail=ObjectSettings("", self.rail_opacity.parsed_value, self.rail_size.parsed_value),
            rail_node=ObjectSettings("", self.rail_node_opacity.parsed_value, self.rail_node_size.parsed_value),
            wall=ObjectSettings(self.wall_color.value, self.wall_opacity.parsed_value, self.wall_size.parsed_value),
            wall_outline=ObjectSettings(self.wall_outline_color.value, self.wall_outline_opacity.parsed_value, 0),
        )

class MapScene(ui.scene):
    def __init__(self, *args, frame_length: int, time_scale: float, zoomout: float=10, **kwargs) -> None:
        kwargs.setdefault("grid", False)
        super().__init__(*args, **kwargs)
        self.time_scale = time_scale
        self.walls: dict[float, tuple[elements.scene_objects.Extrusion, elements.scene_objects.Extrusion]] = {}
        self.wall_lookup: dict[str, float] = {}
        with self:
            self._obj_group = self.group()
            self.move_camera(zoomout,-self.time_scale*zoomout/20,zoomout, 0,self.time_scale/2,0)
            for i in range(frame_length):
                with self.group().move(0,(i+0.5)*self.time_scale,0):
                    self.box(16, self.time_scale, 12, wireframe=True).material("#008", 0.2)
                    self.box(0.1, self.time_scale, 12, wireframe=True).material("#080", 0.1)
                    self.box(16, self.time_scale, 0.1, wireframe=True).material("#800", 0.1)
    @handle_errors
    def _handle_drag(self, e: events.GenericEventArguments) -> None:
        # avoid KeyError when deleting object that is being dragged
        if e.args.get("object_id") not in self.objects:
            return
        super()._handle_drag(e)
    def to_scene(self, xyt: "numpy array (3+)") -> tuple[float, float, float]:
        return (xyt[0], xyt[2]*self.time_scale, xyt[1])
    def _sphere(self, xyt: "numpy array (3+)", obj_settings: ObjectSettings, color: str) -> elements.scene_objects.Sphere:
        return self.sphere(obj_settings.size).move(*self.to_scene(xyt)).material(color, obj_settings.opacity)

    def wall_extrusion(self, xytwa: "numpy array (5)", thickness: float) -> elements.scene_objects.Extrusion:
        return self.extrusion(
            WALL_VERTS[int(xytwa[0,3])].tolist(), -thickness,
        ).rotate(
            np.deg2rad(90), np.deg2rad(180 - xytwa[0,4]), 0
        ).move(
            *self.to_scene(xytwa[0])
        )

    def render(self, data: synth_format.DataContainer, settings: RenderSettings = RenderSettings()) -> None:
        self._obj_group.delete()
        with self.group() as self._obj_group:
            for t, w in data.walls.items():
                body = self.wall_extrusion(
                    w, settings.wall.size * self.time_scale
                ).material(
                    settings.wall.color, settings.wall.opacity
                )

                with body:
                    outline = self.extrusion(
                        WALL_VERTS[int(w[0,3])].tolist(), -settings.wall.size * self.time_scale, wireframe=True
                    ).material(
                        settings.wall_outline.color, settings.wall_outline.opacity
                    )
                self.walls[t] = (body, outline)
                self.wall_lookup[body.id] = t
            for ty in synth_format.NOTE_TYPES:
                color: str = getattr(settings, "color_" + ty)
                for _, n in getattr(data, ty).items():
                    parent = self._sphere(n[0], settings.note, color)
                    with parent:
                        diff = n - n[0]
                        for i in range(1, n.shape[0]):
                            self._sphere(diff[i], settings.rail_node, color)

                            self.quadratic_bezier_tube(
                                list(self.to_scene(diff[i-1])),
                                list(self.to_scene((diff[i-1]+diff[i])/2)),
                                list(self.to_scene(diff[i])),
                                radius=settings.rail.size,
                            ).material(
                                color, settings.rail.opacity
                            )



from typing import Any, Union

import numpy as np

def beat_to_second(beat: float, bpm: float) -> float:
    return beat * 60 / bpm

def second_to_beat(second: float, bpm: float) -> float:
    return second * bpm / 60

def bounded_arange(start: float, end: float, step: float) -> "numpy array (x)":
    # arange would not include the end position, so we intentially overshoot, then correct
    new_times = np.arange(start, end + step, step)
    if new_times.shape[0] > 1 and np.isclose(new_times[-2], end):
        # we overshot by an interval and the penultimate position already closely matches the end -> remove overshoot
        new_times[-2] = end
        return new_times[:-1]
    # we overshot by less than an interval -> clamp last time to end
    new_times[-1] = end
    return new_times

def bounded_arange_plusminus(start: float, end: float, step: float) -> "numpy array (x)":
    # allow step to be negative, start from end then
    if step > 0:
        return bounded_arange(start, end, step)
    else:
        return end - bounded_arange(0, end-start, -step)

def parse_number(val: str) -> float:
    if not val:
        raise ValueError("Value empty")
    if "/" in val:
        num, denom = val.split("/", 1)
        if " " in num:
            # mixed fraction, ie "1 1/2" -> 1.5
            integer, num = num.split(" ", 1)
            i = int(integer)
            return i + np.sign(i) * (float(num) / float(denom))
        return float(num) / float(denom)
    elif val.endswith("%"):
        return float(val[:-1]) / 100
    return float(val)

def parse_range(val: str) -> tuple[float, float]:
    if ":" not in val:
        v = parse_number(val)
        return (-v, v)
    split = val.split(":")
    if len(split) != 2:
        raise ValueError("Must be in the form 'max' or 'min:max'")
    try:
        min = parse_number(split[0])
    except ValueError as ve:
        raise ValueError(f"Error parsing minimum: {ve}") from ve
    try:
        max = parse_number(split[1])
    except ValueError as ve:
        raise ValueError(f"Error parsing maximum: {ve}") from ve
    return (min, max)

def parse_xy_range(val: str) -> tuple[tuple[float, float], tuple[float, float]]:
    split = val.split(",")
    if len(split) != 2:
        raise ValueError("Must be in the form X_RANGE,Y_RANGE")
    try:
        x = parse_range(split[0])
    except ValueError as ve:
        raise ValueError(f"Error parsing x range: {ve}") from ve
    try:
        y = parse_range(split[1])
    except ValueError as ve:
        raise ValueError(f"Error parsing y range: {ve}") from ve
    return ((x[0], y[0]), (x[1], y[1]))


class SecondFloat:
    # dummy class to hold numbers parsed as seconds until bpm was parsed
    def __init__(self, val: float) -> None:
        self.val = val
    def to_beat(self, bpm: float) -> float:
        return second_to_beat(self.val, bpm)
    def __str__(self) -> str:
        return f"{self.val}s"
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"


def parse_time(val: str) -> Union[float, SecondFloat]:
    # note that there is no rounding here,
    if val.endswith("s"):
        return SecondFloat(parse_number(val[:-1]))
    elif ":" in val:
        # parse mm:ss.fff into seconds
        m, s = val.rsplit(":", 1)
        return SecondFloat(float(m) * 60 + float(s))
    return parse_number(val)

def parse_position(val: str) -> tuple[float, float, Union[float, SecondFloat]]:
    split = val.split(",")
    if len(split) != 3:
        raise ValueError("Must be in the form x,y,t")
    try:
        x = parse_number(split[0])
    except ValueError as ve:
        raise ValueError(f"Error parsing x: {ve}") from ve
    try:
        y = parse_number(split[1])
    except ValueError as ve:
        raise ValueError(f"Error parsing y: {ve}") from ve
    try:
        t = parse_time(split[2])
    except ValueError as ve:
        raise ValueError(f"Error parsing t: {ve}") from ve
    return (x, y, t)

def pretty_fraction(val: float) -> str:
    if round(val, 4).is_integer():
        return str(int(round(val, 4)))
    if round(val*192, 4).is_integer():
        if val < 1:
            # regular fraction: a/b
            v = int(round(val*192, 4))
            gcd = np.gcd(v,192)
            return f"{v//gcd}/{192//gcd}"
        else:
            # mixed fraction: i a/b
            i = int(val)
            v = int(round((val-i)*192, 4))
            gcd = np.gcd(v,192)
            return f"{i} {v//gcd}/{192//gcd}"
    return str(val)

def pretty_time_delta(seconds: float) -> str:
    seconds = abs(seconds)
    if seconds < 1:
        return f"{seconds*1000:.0f} ms"
    units = {
        "second": 60,
        "minute": 60,
        "hour": 24,
        "day": 7,
        "week": 30/7,
        "month": 365/30,
        "year": 1000,
    }
    value = seconds
    for unit, next_unit in units.items():
        if value < next_unit:
            # add 's' if value rounds to 2 or more
            return f"{value:.0f} {unit}{'s' if value >= 1.5 else ''}"
        value = value / next_unit
    return "a really long time"


def pretty_list(data: list[Any]) -> str:
    if not data:
        return ""
    if len(data) == 1:
        return str(data[0])
    return ", ".join(map(str, data[:-1])) + f" and {data[-1]}"



import unittest

from synth_mapping_helper.synth_format import import_clipboard, export_clipboard


class TestExportImport(unittest.TestCase):
    def test_reexport(self):
        pass


if __name__ == "__main__":
    unittest.main()



from io import BytesIO
from dataclasses import dataclass, field
from datetime import datetime
from datetime import datetime
from typing import Any, Generator, Literal

import numpy as np
import librosa
import soundfile

from synth_mapping_helper import rails, utils, audio_format
from synth_mapping_helper.synth_format import DataContainer, SINGLE_COLOR_NOTES, NOTE_TYPES, WALL_TYPES, ALL_TYPES, AudioData, GRID_SCALE

RENDER_WINDOW_WALL = 4.0  # the game always renders walls 4 seconds ahead
RENDER_WINDOW_NOTES = 3.5  # the game always renders notes 3.5 seconds ahead
QUEST_WIREFRAME_LIMIT = 200  # combined
QUEST_RENDER_LIMIT = 500  # combined
PC_TYPE_DESPAWN = 80  # for each type

RAIL_NODE_DIST = 2.0  # warn for long rails without intermediate nodes

HEAD_POSITION = np.array([0.0,2.0])
HEAD_RADIUS_SQ = 1.5 ** 2  # pre-squared, such that we can skip the sqrt in sqrt(x**2+y**2) < HEAD_RADIUS

# the game uses some formula like out = in - (in^3) to scale X and Y in spieral, which is not monotonic and breaks down as the coordinates approach 1m
SPIRAL_APEX = 0.65 / GRID_SCALE  # 65 cm away from the center the notes stop getting further out
SPIRAL_FLIP = 0.80 / GRID_SCALE  # 80 cm away from the center the math breaks down completely and returns to center, then flips to other side
SPIRAL_NEUTRAL_OFFSET: dict[str, "numpy array (2,)"] = {
    # center is shifted a bit for left/right hands
    "right": np.array([1.5,0.0]),
    "left": np.array([-1.5,0.0]),
    "single": np.zeros(2),
    "both": np.zeros(2),
}

CURVE_WINDOW_S = 1.0  # time between notes for break in curve
CURVE_INTERP = 192  # 192 bins per beat
RAIL_WEIGHT = 0.3  # rails have a lower weight than notes
RAIL_TAIL_WEIGHT = 0.1  # last 20% of rail
HAND_CURVE_TYPE = tuple["numpy array (n, 2)", "numpy array (n, 2)", "numpy array (n, 2)"]  # curve, velocity, acceleration

END_PADDING = 1.0  # notes should not be in the last second to avoid them not showing up at all

@dataclass
class PlotDataContainer:
    times: list[float]
    plot_data: "numpy array (n, 2)"
    max_value: float = field(init=False)
    plot_times: list[np.datetime64] = field(init=False)

    def __post_init__(self) -> None:
        self.max_value = self.plot_data[:,1].max() if self.plot_data.shape[0] else 0.0
        self.plot_times = [np.datetime64(datetime.utcfromtimestamp(t)) for t in self.plot_data[:,0]]

# DENSITY

def density(times: list[float], window: float) -> PlotDataContainer:
    # prepares density plot
    if not times:
        return PlotDataContainer(times=[], plot_data=np.zeros((0,2)))
    # time, count
    out: list[tuple[float, int]] = []
    visible_t: list[float] = []
    c = 0  # tracks len(visible_t)
    for t in sorted(times):
        start = t - window
        while c and visible_t[0] < start:
            # always create two datapoints to force discrete "steps"
            out.append((visible_t[0], c))
            out.append((visible_t[0], c-1))
            visible_t = visible_t[1:]
            c -= 1
        out.append((start, c))
        out.append((start, c+1))
        visible_t.append(t)
        c += 1
    while visible_t:
        out.append((visible_t[0], c))
        out.append((visible_t[0], c-1))
        visible_t = visible_t[1:]
        c -= 1

    return PlotDataContainer(
        times=times,
        plot_data=np.array(out)
    )

def wall_mode(highest_density: float, *, combined: bool) -> str:
    mode = "OK"
    if combined:
        if highest_density >= QUEST_RENDER_LIMIT:
            mode = "Quest-Limited"
        elif highest_density >= QUEST_WIREFRAME_LIMIT:
            mode = "Quest-Wireframe"
    else:
        if highest_density >= PC_TYPE_DESPAWN:
            mode = "PC-Despawn"

    return f"{mode}, max {round(highest_density)}"

def note_densities(data: DataContainer) -> dict[str, dict[str, PlotDataContainer]]:
    window_b = utils.second_to_beat(RENDER_WINDOW_NOTES, bpm=data.bpm)
    out = {}
    for nt in NOTE_TYPES:
        # time, node_count
        notes = [(t, n.shape[0]) for t, n in getattr(data, nt).items()]
        # time for every single node (excluding rail head)
        all_nodes = [
            xyt[2]
            for t, n in getattr(data, nt).items()
            for xyt in n[1:]
        ]
        out[nt] = {
            "note": density(times=[t for t,_ in notes], window=window_b),
            "single": density(times=[t for t, c in notes if c==1], window=window_b),
            "rail": density(times=[t for t, c in notes if c>1], window=window_b),
            "rail node": density(times=all_nodes, window=window_b),
        }
    out["combined"] = {
        k: density(
            times=[t for nt in NOTE_TYPES for t in out[nt][k].times],
            window=window_b
        )
        for k in out[NOTE_TYPES[0]]
    }
    return out

def all_note_densities(diffs: dict[str, DataContainer]) -> dict[str, dict[str, dict[str, PlotDataContainer]]]:
    return {
        d: note_densities(c)
        for d, c in diffs.items()
    }

def wall_densities(data: DataContainer) -> dict[str, PlotDataContainer]:
    window_b = RENDER_WINDOW_WALL*data.bpm/60
    out = {
        wt: density(times=[t for t, w in data.walls.items() if w[0,3] == tid], window=window_b)
        for wt, (tid, *_) in WALL_TYPES.items()
    }
    out["combined"] = density(times=list(data.walls), window=window_b)
    return out

def all_wall_densities(diffs: dict[str, DataContainer]) -> dict[str, dict[str, PlotDataContainer]]:
    return {
        d: wall_densities(c)
        for d, c in diffs.items()
    }

def running_counts(data: DataContainer) -> dict[str, PlotDataContainer]:
    times: dict[str, list[float]] = {
        t: sorted(data.get_object_dict(t)) for t in ALL_TYPES
    }
    times["notes+rails"] = sorted(ti for t in NOTE_TYPES for ti in times[t])
    times["walls"] = sorted(ti for t in WALL_TYPES for ti in times[t])
    return {
        t: PlotDataContainer(
            times=np.array(tis),
            plot_data=np.array(np.stack((tis, range(1,len(tis)+1)), axis=-1)),
        )
        for t, tis in times.items()
    }

def all_running_counts(diffs: dict[str, DataContainer]) -> dict[str, dict[str, PlotDataContainer]]:
    return {
        d: running_counts(c)
        for d, c in diffs.items()
    }

# MOVEMENT

def hand_curve(notes: SINGLE_COLOR_NOTES, window_b: float) -> HAND_CURVE_TYPE:
    if not notes:
        return np.full((1,3), np.nan), np.full((1,3), np.nan), np.full((1,3), np.nan)
    # step 1: find average positions for each time "bin", by weighted average of notes and (interpolated) rails
    data_points: dict[int, "numpy array (3,)"] = {}  # t//i: weight,x,y
    for _, nodes in sorted(notes.items()):
        tb = int(nodes[0,2]*CURVE_INTERP)  # time bin
        if nodes.shape[0] == 1:  # single notes
            # head at full weight
            wxy = np.concatenate(([1.0], nodes[0,:2]))
            if tb not in data_points:
                data_points[tb] = wxy
            else:
                data_points[tb] += wxy
        else:  # rails
            # first, sample at 1/192
            interp_rail = rails.interpolate_nodes(nodes, mode="spline", interval=1/CURVE_INTERP)[:,:2]
            # then, add weights: full for head, reduced for tails
            weights = np.full((interp_rail.shape[0], 1), RAIL_WEIGHT)
            weights[0] = 1.0
            weights[-int(interp_rail.shape[0]*0.2):] = RAIL_TAIL_WEIGHT
            weighted_rail = np.concatenate((weights, interp_rail*weights), axis=-1)
            # finally, dump bins into data points
            for bi, wxy in enumerate(weighted_rail):
                if tb+bi not in data_points:
                    data_points[tb+bi] = wxy
                else:
                    data_points[tb+bi] += wxy
    # step 2: locate continguous sections and interpolate over the averaged locations
    out_rails = []
    out_curve = []
    out_vel = []
    out_acc = []
    current_curve: list[tuple[float, float, float]] = []

    def _append_section():
        interp_pos = rails.interpolate_nodes(np.array(current_curve), mode="hermite", interval=1/CURVE_INTERP)
        out_curve.append(interp_pos)
        out_curve.append(np.full((1,3), np.nan))  # add NaN spacers between sections

        if interp_pos.shape[0] > 1:
            section_vel = np.diff(interp_pos[:, :2], axis=0)
            out_vel.append(np.concatenate((section_vel, interp_pos[:-1, 2:3]), axis=-1))
            out_vel.append(np.full((1,3), np.nan))
            if section_vel.shape[0] > 1:
                section_acc = np.diff(section_vel, axis=0)
                out_acc.append(np.concatenate((section_acc, interp_pos[1:-1, 2:3]), axis=-1))
                out_acc.append(np.full((1,3), np.nan))
        
        current_curve.clear()

    last_b = None
    for bi, (w,x,y) in sorted(data_points.items()):
        this_b = (bi/CURVE_INTERP)
        if last_b is not None and (this_b - last_b) > window_b:
            _append_section()
        current_curve.append((x/w,y/w,this_b))
        last_b = this_b
    if current_curve:
        _append_section()
    # put it all together
    return (
        np.concatenate(out_curve),
        np.concatenate(out_vel) if out_vel else np.full((1,3), np.nan),
        np.concatenate(out_acc) if out_acc else np.full((1,3), np.nan)
    )

def hand_curves(data: DataContainer) -> dict[str, HAND_CURVE_TYPE]:
    return {
        nt: hand_curve(getattr(data, nt), window_b=utils.second_to_beat(CURVE_WINDOW_S, bpm=data.bpm))
        for nt in NOTE_TYPES
        if getattr(data, nt)
    }

def all_hand_curves(diffs: dict[str, DataContainer]) -> dict[str, dict[str, HAND_CURVE_TYPE]]:
    return {
        d: hand_curves(c)
        for d, c in diffs.items()
    }

def sections_from_bools(bools: "numpy array (n,)") -> Generator[tuple[int, int], None, None]:
    if bools.any():
        idx = np.nonzero(bools)[0]
        idx_diff_idx = np.nonzero(np.diff(idx)-1)[0]
        start = 0
        for idi in idx_diff_idx:
            yield idx[start], idx[idi]
            start = idi + 1
        yield idx[start], idx[-1]


# type: icon, text
WARNING_TYPES: dict[str, tuple[str, str]] = {
    "end_padding": ("🔚", "Close to end of song when accounting for offset.<br>This may cause issues."),
    "straight_rail": ("📏", "Long rail without intermediate nodes.<br>This may be unplayable in spiral."),
    "spiral_distortion": ("🌀<br>⚠️", "Somewhat far from hand neutral position.<br>This may look distorted in spiral."),
    "spiral_breakdown": ("🌀<br>💀", "Far away from hand neutral position.<br>This will be massively misplaced in spiral."),
    "head_area": ("🙈" , "Inside head area.<br>This may block line of sight")
}

@dataclass
class Warning:
    type: Literal["straight_rail", "end_padding", "spiral_distortion", "spiral_breakdown", "head_area"]
    figure: Literal["x", "y", "xy"]
    note_type: Literal["left", "right", "single", "both"]
    note_rail: Literal["note", "rail"]
    start_beat: float
    end_beat: float

    @property
    def icon(self) -> str:
        return WARNING_TYPES[self.type][0]
    
    @property
    def text(self) -> str:
        return WARNING_TYPES[self.type][1]

def warnings(data: DataContainer, last_beat: float) -> list[Warning]:
    last_safe_beat = last_beat - utils.second_to_beat(END_PADDING, bpm=data.bpm)
    out: list[Warning] = []
    for nt in NOTE_TYPES:
        for _, nodes in sorted(getattr(data, nt).items()):
            if nodes.shape[0] == 1:
                crv = nodes
                nr = "note"
            else:
                crv = rails.interpolate_nodes(nodes, mode="spline", interval=1/CURVE_INTERP)
                nr = "rail"

                rail_deltas = np.diff(nodes[:,2])
                # nodes more than X beats apart
                for s_idx, e_idx in sections_from_bools(rail_deltas > RAIL_NODE_DIST):
                    out.append(Warning(
                        type="straight_rail",
                        figure="xy",
                        note_type=nt,
                        note_rail=nr,
                        start_beat=crv[s_idx, 2],
                        end_beat=crv[e_idx, 2],
                    ))
            # notes/rails near end
            if nodes[-1, 2] >= last_safe_beat:
                out.append(Warning(
                    type="end_padding",
                    figure="xy",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=max(nodes[0, 2], last_safe_beat),
                    end_beat=nodes[-1, 2],
                ))

            spiral_delta = np.abs((crv[:,:1] - SPIRAL_NEUTRAL_OFFSET[nt]))
            # delta x between apex and flip
            for s_idx, e_idx in sections_from_bools(np.logical_and(spiral_delta[:,0] > SPIRAL_APEX, spiral_delta[:,0] <= SPIRAL_FLIP)):
                out.append(Warning(
                    type="spiral_distortion",
                    figure="x",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=crv[s_idx, 2],
                    end_beat=crv[e_idx, 2],
                ))
            # delta y between apex and flip
            for s_idx, e_idx in sections_from_bools(np.logical_and(spiral_delta[:,1] > SPIRAL_APEX, spiral_delta[:,1] <= SPIRAL_FLIP)):
                out.append(Warning(
                    type="spiral_distortion",
                    figure="y",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=crv[s_idx, 2],
                    end_beat=crv[e_idx, 2],
                ))
            # delta x beyond flip
            for s_idx, e_idx in sections_from_bools(spiral_delta[:,0] > SPIRAL_FLIP):
                out.append(Warning(
                    type="spiral_breakdown",
                    figure="x",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=crv[s_idx, 2],
                    end_beat=crv[e_idx, 2],
                ))
            # delta x beyond flip
            for s_idx, e_idx in sections_from_bools(spiral_delta[:,1] > SPIRAL_FLIP):
                out.append(Warning(
                    type="spiral_breakdown",
                    figure="y",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=crv[s_idx, 2],
                    end_beat=crv[e_idx, 2],
                ))

            head_delta = (crv[:,:2] - HEAD_POSITION)
            # distance to head less than keepout radius
            for s_idx, e_idx in sections_from_bools(head_delta[:,0]**2+head_delta[:,1]**2 <= HEAD_RADIUS_SQ):
                out.append(Warning(
                    type="head_area",
                    figure="xy",
                    note_type=nt,
                    note_rail=nr,
                    start_beat=crv[s_idx, 2],
                    end_beat=crv[e_idx, 2],
                ))
    return sorted(out, key=lambda w: w.start_beat)

def all_warnings(diffs: dict[str, DataContainer], last_beat: float) -> dict[str, list[tuple[str, str, float, float, str]]:]:
    return {
        d: warnings(c, last_beat=last_beat)
        for d, c in diffs.items()
    }

# AUDIO

def calculate_onsets(data: "numpy array (s,)", sr: int) -> "numpy array (m,)":
    return librosa.util.normalize(librosa.onset.onset_strength(y=data, sr=sr, aggregate=np.median, center=True))

def find_bpm(onsets: "numpy array (m,)", sr: int) -> "numpy array (m,), numpy array (m,), numpy array (m,)":
    # bins between 0 and approximately 300 bpm (not sure why, but this formula works out)
    hop_len = 1<<(sr.bit_length()-4)
    # decrease hop for shorter signals
    while hop_len * 3 // 2 > onsets.shape[-1]:
        hop_len //= 2
    # 50 % overlap
    win_len = hop_len * 3 // 2
    # this is based on librosa.beat.plp
    ftgram = librosa.feature.fourier_tempogram(onset_envelope=onsets, sr=sr, hop_length=hop_len, win_length=win_len)
    tempo_frequencies = librosa.fourier_tempo_frequencies(sr=sr, hop_length=hop_len, win_length=win_len)
    ftgram[..., tempo_frequencies < 60, :] = 0
    ftgram[..., tempo_frequencies > 240, :] = 0
    ftmag = np.log1p(1e6 * np.abs(ftgram))
    peak_freq_bins = ftmag.argmax(axis=-2)
    # Now we have peak bins, estimate the binning error (±1/2) using
    # https://ccrma.stanford.edu/~jos/sasp/Quadratic_Interpolation_Spectral_Peaks.html
    # Then we can store BPM and peak value for every frame
    bpm_peaks = []
    bpm_peak_values = []
    bpm_multiplier = sr/(hop_len*win_len) * 60  # used to convert intermediate bin to bpm
    for frame, freq_bin in enumerate(peak_freq_bins):
        if freq_bin == 0 or freq_bin == ftgram.shape[-1]:
            bpm_peaks.append(0)
            bpm_peak_values.append(0)
        else:
            a, b, c = np.abs(ftgram[..., freq_bin-1:freq_bin+2, frame])
            p = 1/2 * (a-c) / (a-2*b+c)
            bpm_peaks.append((freq_bin+p)*bpm_multiplier)
            bpm_peak_values.append(b-1/4*(a-c)*p)
    # back to plp
    peak_values = ftmag.max(axis=-2, keepdims=True)
    ftgram[ftmag < peak_values] = 0
    ftgram /= librosa.util.tiny(ftgram) ** 0.5 + np.abs(ftgram.max(axis=-2, keepdims=True))
    pulse = librosa.istft(ftgram, hop_length=1, n_fft=win_len, length=onsets.shape[-1])
    pulse = np.clip(pulse, 0, None, pulse)
    # bpms, normalized bpm strength, pulse curve
    return np.array(bpm_peaks), librosa.util.normalize(np.array(bpm_peak_values)), librosa.util.normalize(pulse)

def group_bpm(bpms: "numpy array (m,)", bpm_strengths: "numpy array (m,)", max_jump: float=0.1, min_len_ratio: float=0.01) -> tuple[float, list[tuple[int, int, float, float]]]:
    min_len = np.ceil(bpms.shape[-1] * min_len_ratio)
    jumps = [idx[0] for idx in np.argwhere(np.abs(np.diff(bpms, prepend=bpms[0])) > max_jump)]

    out: list[tuple[int, int, float, float]] = []
    max_str = 0
    best_bpm = 0
    for start, end in zip([0, *jumps], [*jumps, bpms.shape[-1]]):
        if (end - start) < min_len:
            # ignore short sections
            continue
        bpm = round(bpms[start:end].mean(), 1)  # round output to 1 decimal
        str_sum = bpm_strengths[start:end].sum()
        if str_sum > max_str:
            max_str = str_sum
            best_bpm = bpm
        out.append((start, end, bpm, str_sum))
    # normalize strength
    return best_bpm, [(s,e,b,st/max_str) for s,e,b,st in out]

def locate_beats(onsets: "numpy array (m,)", sr: int, bpm: float) -> "numpy array (t,)":
    _, beats = librosa.beat.beat_track(onset_envelope=onsets, bpm=bpm, sr=sr, trim=False)
    return beats

def circmedian(values: "numpy array (n,)", high: float) -> float:
    # doing statistics on "circular data" (ie 0-beat_time) is hard, but we can treat each value as "angle" (0-2pi, ie 0-360 deg)
    # see also: scipy.stats.circmean
    # via median of sine and cosine, we get the "median angle" and transform that back
    # we get the median instead of mean, to avoid outliers influencing the result
    scaling = 2*np.pi/high
    sines = np.sin(values*scaling)
    cosines = np.cos(values*scaling)
    return (np.arctan2(np.median(sines), np.median(cosines)) / scaling + high)%high

def circerror(values: "numpy array (n,)", target: float, high: float) -> "numpy array (n,)":
    # shift the delta such that equal -> h/2  and opposite -> 0 or h
    shifted_delta = (values - target + high*1.5) % high
    # now get the delta from high/2, and transform into 0 (equal) to 1 (opposite)
    return np.abs(shifted_delta-high/2) / (high/2)

def find_offsets(onsets: "numpy array (m,)", sr: int, bpm_sections: list[tuple[int, int, float, float]]) -> list[tuple[int, int, "numpy array (n,)", float, int, "numpy array (n,)"]]:
    offset_sections = []
    for i, (section_start, section_end, section_bpm, _) in enumerate(bpm_sections):
        beats = locate_beats(onsets=onsets[section_start:section_end], sr=sr, bpm=section_bpm)
        if not beats.any():
            # ignore sections without detected beats
            continue
        # not sure why, but a 22ms offset seems to be required...
        beats = librosa.frames_to_time(beats + section_start, sr=sr) - 0.022
        beat_time = 60/section_bpm
        median_offset = circmedian(beats % beat_time, high=beat_time)
        offset_error = circerror(beats % beat_time, median_offset, high=beat_time)
        offset_ms = int((beat_time - median_offset)*1000)  # the game offsets the audio, so negate offset
        offset_sections.append((section_start, section_end, beats, section_bpm, offset_ms, offset_error))
    return offset_sections

def bpm_aio(raw_data: bytes) -> dict[str, Any]:
    data, sr = audio_format.load_for_analysis(raw_data)
    onsets = calculate_onsets(data=data, sr=sr)
    peak_bpms, peak_values, pulse = find_bpm(onsets=onsets, sr=sr)
    best_bpm, bpm_sections = group_bpm(bpms=peak_bpms, bpm_strengths=peak_values)
    offset_sections = find_offsets(onsets=onsets, sr=sr, bpm_sections=bpm_sections)
    return {
        "data": data,
        "sr": sr,
        "onsets": onsets,
        "peak_bpms": peak_bpms,
        "peak_values": peak_values,
        "pulse": pulse,
        "best_bpm": best_bpm,
        "bpm_sections": bpm_sections,
        "offset_sections": offset_sections,
    }